##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import unittest

from Acquisition import Implicit
from conformance import IBasicUser_conformance, \
                        IPropertiedUser_conformance

class FauxMethod:

    def __init__( self, im_self, local_roles=() ):

        self.im_self = im_self
        self.__ac_local_roles__ = local_roles

class FauxProtected( Implicit ):

    def __init__( self, local_roles=() ):

        self.__ac_local_roles__ = local_roles

class PropertiedUserTests( unittest.TestCase
                           , IBasicUser_conformance
                           , IPropertiedUser_conformance
                           ):

    def _getTargetClass( self ):

        from Products.PluggableAuthService.PropertiedUser \
            import PropertiedUser

        return PropertiedUser

    def _makeOne( self, id='testing', login=None, *args, **kw ):

        return self._getTargetClass()( id, login, *args, **kw )

    def test_empty( self ):

        user = self._makeOne( 'empty' )

        # BaseUser interface
        self.assertEqual( user.getId(), 'empty' )
        self.assertEqual( user.getUserName(), 'empty' )
        self.assertRaises( NotImplementedError, user._getPassword )
        self.assertEqual( len( user.getRoles() ), 0 )
        self.assertEqual(len(user.getGroups()), 0)
        self.assertEqual( len( user.getDomains() ), 0 )

        # plus propertysheets
        self.assertEqual( len( user.listPropertysheets() ), 0 )
        self.assertRaises( KeyError, user.getPropertysheet, 'nonesuch'  )

    def test_groups(self):
        groups = ('Group A', 'Group B')
        user = self._makeOne('groups')
        user._addGroups(groups)
        self.assertEqual(len(user.getGroups()), len(groups))
        for g in user.getGroups():
            self.assert_(g in groups)

    def test_username(self):
        user = self._makeOne('username', 'User with Username')
        self.assertEqual(user.getUserName(), 'User with Username')

    def test_roles(self):
        roles = ['Manager', 'Members']
        user = self._makeOne('user')
        user._addRoles(roles)
        self.assertEqual(len(user.getRoles()), 2)
        for r in user.getRoles():
            self.assert_(r in roles)

    def test_addPropertysheet( self ):

        user = self._makeOne()

        user.addPropertysheet( 'one', { 'a' : 0, 'b' : 'jabber' } )

        ids = user.listPropertysheets()

        self.assertEqual( len( ids ), 1 )
        self.assertEqual( ids[ 0 ], 'one' )

        sheet = user.getPropertysheet( 'one' )

        self.assertEqual( len( sheet.propertyMap() ), 2 )
        self.assertEqual( sheet.getPropertyType( 'a' ), 'int' )
        self.assertEqual( sheet.getPropertyType( 'b' ), 'string' )
        self.assertEqual( sheet.getId(), 'one' )

        sheet = user[ 'one' ]

        self.assertEqual( len( sheet.propertyMap() ), 2 )
        self.assertEqual( sheet.getPropertyType( 'a' ), 'int' )
        self.assertEqual( sheet.getPropertyType( 'b' ), 'string' )

        self.assertRaises( KeyError, user.getPropertysheet, 'another'  )

    def test_getRolesInContext_no_local( self ):

        groups = ( 'Group A', 'Group B' )
        user = self._makeOne()
        user._addGroups( groups )

        self.assertEqual( len( user.getRoles() ), 0 )
        self.assertEqual( len( user.getGroups() ), len( groups ) )

        faux = FauxProtected()

        local_roles = user.getRolesInContext( faux )
        self.assertEqual( len( local_roles ), 0 )

    def test_getRolesInContext_group_match( self ):

        groups = ( 'Group A', 'Group B' )
        user = self._makeOne()
        user._addGroups( groups )

        faux = FauxProtected( { 'Group A' : ( 'Manager', ) } )

        local_roles = user.getRolesInContext( faux )
        self.assertEqual( len( local_roles ), 1 )
        self.failUnless( 'Manager' in local_roles )

    def test_getRolesInContext_group_overlap( self ):

        groups = ( 'Group A', 'Group B' )
        user = self._makeOne()
        user._addGroups( groups )

        faux = FauxProtected( { 'Group A' : ( 'Manager', )
                              , 'Group B' : ( 'Manager', 'Owner' )
                              } )

        local_roles = user.getRolesInContext( faux )
        self.assertEqual( len( local_roles ), 2 )
        self.failUnless( 'Manager' in local_roles )
        self.failUnless( 'Owner' in local_roles )

    def test_getRolesInContext_group_nomatch( self ):

        groups = ( 'Group A', 'Group B' )
        user = self._makeOne()
        user._addGroups( groups )

        faux = FauxProtected( { 'Group C' : ( 'Manager', ) } )

        local_roles = user.getRolesInContext( faux )
        self.assertEqual( len( local_roles ), 0 )

    def test_getRolesInContext_acquisition( self ):

        groups = ( 'Group A', 'Group B' )
        user = self._makeOne()
        user._addGroups( groups )

        faux_container = FauxProtected( { 'Group A' : ( 'Manager', ) } )
        faux_contained = FauxProtected( { 'Group C' : ( 'Manager', 'Owner' ) }
                                      ).__of__( faux_container )

        local_roles = user.getRolesInContext( faux_contained )
        self.assertEqual( len( local_roles ), 1 )
        self.failUnless( 'Manager' in local_roles )

    def test_getRolesInContext_weslayan( self ):

        # Test "methodish" checks.

        groups = ( 'Group A', 'Group B' )
        user = self._makeOne()
        user._addGroups( groups )

        faux_self = FauxProtected( { 'Group A' : ( 'Manager', ) } )
        faux_method = FauxMethod( faux_self
                                , { 'Group C' : ( 'Manager', 'Owner' ) } )

        local_roles = user.getRolesInContext( faux_method )
        self.assertEqual( len( local_roles ), 1 )
        self.failUnless( 'Manager' in local_roles )

    def test_allowed_not_even_god_should( self ):

        from AccessControl.PermissionRole import _what_not_even_god_should_do
        user = self._makeOne()

        self.failIf( user.allowed( None, _what_not_even_god_should_do ) )

    def test_allowed_anonymous( self ):

        user = self._makeOne()

        self.failUnless( user.allowed( None, ('Anonymous',) ) )

    def test_allowed_authenticated( self ):

        user = self._makeOne()

        self.failUnless( user.allowed( None, ('Authenticated',) ) )

    def test_allowed_authenticated_required_but_anonymous( self ):

        user = self._makeOne('Anonymous User')

        self.failIf( user.allowed( None, ('Authenticated',) ) )

    def test_allowed_global_roles_ok( self ):

        user = self._makeOne()
        user._addRoles( ( 'Role 1', 'Role 2' ) )

        self.failUnless( user.allowed( None, ( 'Role 1', ) ) )

    def test_allowed_global_roles_not_ok( self ):

        user = self._makeOne()
        user._addRoles( ( 'Role 1', 'Role 2' ) )

        self.failIf( user.allowed( None, ( 'Role 3', ) ) )

    def test_allowed_local_roles_on_user_ok( self ):

        user = self._makeOne( 'user' )
        object = FauxProtected( { 'user' : ( 'Role 1', ) } )

        self.failUnless( user.allowed( object, ( 'Role 1', ) ) )

    def test_allowed_local_roles_on_user_not_ok( self ):

        user = self._makeOne( 'user' )
        object = FauxProtected( { 'user' : ( 'Role 1', ) } )

        self.failIf( user.allowed( object, ( 'Role 2', ) ) )

    def test_allowed_local_roles_on_group_ok( self ):

        user = self._makeOne( 'user' )
        user._addGroups( ( 'Group 1', 'Group 2' ) )
        object = FauxProtected( { 'Group 1' : ( 'Role 1', ) } )

        self.failUnless( user.allowed( object, ( 'Role 1', ) ) )

    def test_allowed_acquisition( self ):

        groups = ( 'Group A', 'Group B' )
        user = self._makeOne()
        user._addGroups( groups )

        faux_container = FauxProtected( { 'Group A' : ( 'Manager', ) } )
        faux_contained = FauxProtected( { 'Group C' : ( 'Manager', 'Owner' ) }
                                      ).__of__( faux_container )

        self.failUnless( user.allowed( faux_contained, ( 'Manager', ) ) )

    def test_allowed_weslayan( self ):

        # Test "methodish" checks.

        groups = ( 'Group A', 'Group B' )
        user = self._makeOne()
        user._addGroups( groups )

        faux_self = FauxProtected( { 'Group A' : ( 'Manager', ) } )
        faux_method = FauxMethod( faux_self
                                , { 'Group C' : ( 'Manager', 'Owner' ) } )

        self.failUnless( user.allowed( faux_method, ( 'Manager', ) ) )


if __name__ == "__main__":
    unittest.main()

def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite( PropertiedUserTests ),
        ))




##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Interfaces for PluggableAuthService

$Id$
"""
# Eventually, those interfaces should be moved to Zope and imported
# here as aliases for backward compatibility.

from zope.interface import Interface


class IRequest( Interface ):

    """ Base Request Interface

    XXX Add methods from BaseRequest?
    """

class IHTTPRequest( IRequest ):

    """ HTTP Request
    """

class IBrowserRequest( IHTTPRequest ):

    """Browser Request
    """

class IWebDAVRequest( IHTTPRequest ):

    """ WebDAV Request
    """

class IXMLRPCRequest( IHTTPRequest ):

    """ XML-RPC Request
    """

class IFTPRequest( IRequest ):

    """ FTP Request
    """




##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest
from Testing import ZopeTestCase

ZopeTestCase.installProduct('PythonScripts')
ZopeTestCase.installProduct('PluginRegistry')
ZopeTestCase.installProduct('PluggableAuthService')
ZopeTestCase.installProduct('GenericSetup')

def test_suite():
    suite = unittest.TestSuite()
    package = 'Products.PluggableAuthService.plugins.tests'
    tests = [
        ZopeTestCase.FunctionalDocFileSuite('ChallengeProtocolChooser.txt',
                                            package=package),
        ]
    for t in tests:
        suite.addTest(t)
    return suite



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Class: CookieAuthHelper

$Id$
"""

from base64 import encodestring, decodestring
from binascii import Error
from urllib import quote, unquote

from AccessControl.SecurityInfo import ClassSecurityInfo
from AccessControl.Permissions import view
from OFS.Folder import Folder
from App.class_init import default__class_init__ as InitializeClass

from zope.interface import Interface

from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate

from Products.PluggableAuthService.interfaces.plugins import \
        ILoginPasswordHostExtractionPlugin
from Products.PluggableAuthService.interfaces.plugins import \
        IChallengePlugin
from Products.PluggableAuthService.interfaces.plugins import \
        ICredentialsUpdatePlugin
from Products.PluggableAuthService.interfaces.plugins import \
        ICredentialsResetPlugin
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.utils import classImplements


class ICookieAuthHelper(Interface):
    """ Marker interface.
    """

manage_addCookieAuthHelperForm = PageTemplateFile(
    'www/caAdd', globals(), __name__='manage_addCookieAuthHelperForm')


def addCookieAuthHelper( dispatcher
                       , id
                       , title=None
                       , cookie_name=''
                       , REQUEST=None
                       ):
    """ Add a Cookie Auth Helper to a Pluggable Auth Service. """
    sp = CookieAuthHelper(id, title, cookie_name)
    dispatcher._setObject(sp.getId(), sp)

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect( '%s/manage_workspace'
                                      '?manage_tabs_message='
                                      'CookieAuthHelper+added.'
                                    % dispatcher.absolute_url() )


class CookieAuthHelper(Folder, BasePlugin):
    """ Multi-plugin for managing details of Cookie Authentication. """

    meta_type = 'Cookie Auth Helper'
    cookie_name = '__ginger_snap'
    login_path = 'login_form'
    security = ClassSecurityInfo()

    _properties = ( { 'id'    : 'title'
                    , 'label' : 'Title'
                    , 'type'  : 'string'
                    , 'mode'  : 'w'
                    }
                  , { 'id'    : 'cookie_name'
                    , 'label' : 'Cookie Name'
                    , 'type'  : 'string'
                    , 'mode'  : 'w'
                    }
                  , { 'id'    : 'login_path'
                    , 'label' : 'Login Form'
                    , 'type'  : 'string'
                    , 'mode'  : 'w'
                    }
                  )

    manage_options = ( BasePlugin.manage_options[:1]
                     + Folder.manage_options[:1]
                     + Folder.manage_options[2:]
                     )

    def __init__(self, id, title=None, cookie_name=''):
        self._setId(id)
        self.title = title

        if cookie_name:
            self.cookie_name = cookie_name


    security.declarePrivate('extractCredentials')
    def extractCredentials(self, request):
        """ Extract credentials from cookie or 'request'. """
        creds = {}
        cookie = request.get(self.cookie_name, '')
        # Look in the request.form for the names coming from the login form
        login = request.form.get('__ac_name', '')

        if login and request.form.has_key('__ac_password'):
            creds['login'] = login
            creds['password'] = request.form.get('__ac_password', '')

        elif cookie and cookie != 'deleted':
            raw = unquote(cookie)
            try:
                cookie_val = decodestring(raw)
            except Error:
                # Cookie is in a different format, so it is not ours
                return creds

            try:
                login, password = cookie_val.split(':')
            except ValueError:
                # Cookie is in a different format, so it is not ours
                return creds

            try:
                creds['login'] = login.decode('hex')
                creds['password'] = password.decode('hex')
            except TypeError:
                # Cookie is in a different format, so it is not ours
                return {}

        if creds:
            creds['remote_host'] = request.get('REMOTE_HOST', '')

            try:
                creds['remote_address'] = request.getClientAddr()
            except AttributeError:
                creds['remote_address'] = request.get('REMOTE_ADDR', '')

        return creds


    security.declarePrivate('challenge')
    def challenge(self, request, response, **kw):
        """ Challenge the user for credentials. """
        return self.unauthorized()


    security.declarePrivate('updateCredentials')
    def updateCredentials(self, request, response, login, new_password):
        """ Respond to change of credentials (NOOP for basic auth). """
        cookie_str = '%s:%s' % (login.encode('hex'), new_password.encode('hex'))
        cookie_val = encodestring(cookie_str)
        cookie_val = cookie_val.rstrip()
        response.setCookie(self.cookie_name, quote(cookie_val), path='/')


    security.declarePrivate('resetCredentials')
    def resetCredentials(self, request, response):
        """ Raise unauthorized to tell browser to clear credentials. """
        response.expireCookie(self.cookie_name, path='/')


    security.declarePrivate('manage_afterAdd')
    def manage_afterAdd(self, item, container):
        """ Setup tasks upon instantiation """
        if not 'login_form' in self.objectIds():
            login_form = ZopePageTemplate( id='login_form'
                                           , text=BASIC_LOGIN_FORM
                                           )
            login_form.title = 'Login Form'
            login_form.manage_permission(view, roles=['Anonymous'], acquire=1)
            self._setObject( 'login_form', login_form, set_owner=0 )


    security.declarePrivate('unauthorized')
    def unauthorized(self):
        req = self.REQUEST
        resp = req['RESPONSE']

        # If we set the auth cookie before, delete it now.
        if resp.cookies.has_key(self.cookie_name):
            del resp.cookies[self.cookie_name]

        # Redirect if desired.
        url = self.getLoginURL()
        if url is not None:
            came_from = req.get('came_from', None)

            if came_from is None:
                came_from = req.get('ACTUAL_URL', '')
                query = req.get('QUERY_STRING')
                if query:
                    if not query.startswith('?'):
                        query = '?' + query
                    came_from = came_from + query
            else:
                # If came_from contains a value it means the user
                # must be coming through here a second time
                # Reasons could be typos when providing credentials
                # or a redirect loop (see below)
                req_url = req.get('ACTUAL_URL', '')

                if req_url and req_url == url:
                    # Oops... The login_form cannot be reached by the user -
                    # it might be protected itself due to misconfiguration -
                    # the only sane thing to do is to give up because we are
                    # in an endless redirect loop.
                    return 0

            if '?' in url:
                sep = '&'
            else:
                sep = '?'
            url = '%s%scame_from=%s' % (url, sep, quote(came_from))
            resp.redirect(url, lock=1)
            resp.setHeader('Expires', 'Sat, 01 Jan 2000 00:00:00 GMT')
            resp.setHeader('Cache-Control', 'no-cache')
            return 1

        # Could not challenge.
        return 0


    security.declarePrivate('getLoginURL')
    def getLoginURL(self):
        """ Where to send people for logging in """
        if self.login_path.startswith('/') or '://' in self.login_path:
            return self.login_path
        elif self.login_path != '':
            return '%s/%s' % (self.absolute_url(), self.login_path)
        else:
            return None

    security.declarePublic('login')
    def login(self):
        """ Set a cookie and redirect to the url that we tried to
        authenticate against originally.
        """
        request = self.REQUEST
        response = request['RESPONSE']

        login = request.get('__ac_name', '')
        password = request.get('__ac_password', '')

        # In order to use the CookieAuthHelper for its nice login page
        # facility but store and manage credentials somewhere else we need
        # to make sure that upon login only plugins activated as
        # IUpdateCredentialPlugins get their updateCredentials method
        # called. If the method is called on the CookieAuthHelper it will
        # simply set its own auth cookie, to the exclusion of any other
        # plugins that might want to store the credentials.
        pas_instance = self._getPAS()

        if pas_instance is not None:
            pas_instance.updateCredentials(request, response, login, password)

        came_from = request.form['came_from']

        return response.redirect(came_from)

classImplements( CookieAuthHelper
               , ICookieAuthHelper
               , ILoginPasswordHostExtractionPlugin
               , IChallengePlugin
               , ICredentialsUpdatePlugin
               , ICredentialsResetPlugin
               )

InitializeClass(CookieAuthHelper)


BASIC_LOGIN_FORM = """<html>
  <head>
    <title> Login Form </title>
  </head>

  <body>

    <h3> Please log in </h3>

    <form method="post" action=""
          tal:attributes="action string:${here/absolute_url}/login">

      <input type="hidden" name="came_from" value=""
             tal:attributes="value request/came_from | string:"/>
      <table cellpadding="2">
        <tr>
          <td><b>Login:</b> </td>
          <td><input type="text" name="__ac_name" size="30" /></td>
        </tr>
        <tr>
          <td><b>Password:</b></td>
          <td><input type="password" name="__ac_password" size="30" /></td>
        </tr>
        <tr>
          <td colspan="2">
            <br />
            <input type="submit" value=" Log In " />
          </td>
        </tr>
      </table>

    </form>

  </body>

</html>
"""




##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Class: BasePlugin

$Id$
"""
from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager
from Acquisition import aq_base, aq_parent, aq_inner
from AccessControl import ClassSecurityInfo
from App.class_init import default__class_init__ as InitializeClass

from zope.interface import implementedBy
from zope.interface import providedBy

from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from Products.PluggableAuthService.utils import classImplements
from Products.PluggableAuthService.permissions import ManageUsers
from Products.PluggableAuthService.utils import createViewName

def flattenInterfaces(implemented):
    return implemented.flattened()


class BasePlugin(SimpleItem, PropertyManager):

    """ Base class for all PluggableAuthService Plugins
    """

    security = ClassSecurityInfo()

    manage_options = ( ( { 'label': 'Activate',
                           'action': 'manage_activateInterfacesForm', }
                         ,
                       )
                     + SimpleItem.manage_options
                     + PropertyManager.manage_options
                     )

    prefix = ''

    _properties = (
        dict(id='prefix', type='string', mode='w',
             label='Optional Prefix'),)

    security.declareProtected( ManageUsers, 'manage_activateInterfacesForm' )
    manage_activateInterfacesForm = PageTemplateFile(
        'www/bpActivateInterfaces', globals(),
        __name__='manage_activateInterfacesForm')

    security.declareProtected( ManageUsers, 'listInterfaces' )
    def listInterfaces( self ):
        """ For ZMI update of interfaces. """

        results = []

        for iface in flattenInterfaces( providedBy( self ) ):
            results.append( iface.__name__ )

        return results

    security.declareProtected( ManageUsers, 'testImplements' )
    def testImplements( self, interface ):
        """ Can't access Interface.providedBy() directly in ZPT. """
        return interface.providedBy( self )

    security.declareProtected( ManageUsers, 'manage_activateInterfaces' )
    def manage_activateInterfaces( self, interfaces, RESPONSE=None ):
        """ For ZMI update of active interfaces. """

        pas_instance = self._getPAS()
        plugins = pas_instance._getOb( 'plugins' )

        active_interfaces = []

        for iface_name in interfaces:
            active_interfaces.append( plugins._getInterfaceFromName(
                                                iface_name ) )

        pt = plugins._plugin_types
        id = self.getId()

        for type in pt:
            ids = plugins.listPluginIds( type )
            if id not in ids and type in active_interfaces:
                plugins.activatePlugin( type, id ) # turn us on
            elif id in ids and type not in active_interfaces:
                plugins.deactivatePlugin( type, id ) # turn us off

        if RESPONSE is not None:
            RESPONSE.redirect('%s/manage_workspace'
                              '?manage_tabs_message='
                              'Interface+activations+updated.'
                            % self.absolute_url())

    security.declarePrivate( '_getPAS' )
    def _getPAS( self ):
        """ Canonical way to get at the PAS instance from a plugin """
        return aq_parent( aq_inner( self ) )

    security.declarePrivate( '_invalidatePrincipalCache' )
    def _invalidatePrincipalCache( self, id ):
        pas = self._getPAS()
        if pas is not None and hasattr( aq_base(pas), 'ZCacheable_invalidate'):
            view_name = createViewName('_findUser', id)
            pas.ZCacheable_invalidate(view_name)

    security.declarePublic( 'applyTransform' )
    def applyTransform( self, value ):
        """ Transform for login name.

        Possibly transform the login, for example by making it lower
        case.

        Normally this is done in PAS itself, but in some cases a
        method in a plugin may need to do it itself, when there is no
        method in PAS that calls that method.
        """
        pas = self._getPAS()
        if pas is not None:
            return pas.applyTransform(value)
        return value


classImplements(BasePlugin, *implementedBy(SimpleItem))

InitializeClass(BasePlugin)



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Base classes for testing plugin interface conformance.

$Id$
"""

from zope.interface.verify import verifyClass


class IExtractionPlugin_conformance:

    def test_IExtractionPlugin_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import IExtractionPlugin

        verifyClass( IExtractionPlugin, self._getTargetClass() )

    def test_IExtractionPlugin_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import IExtractionPlugin

        listed = self._makeOne().listInterfaces()
        self.failUnless(IExtractionPlugin.__name__ in listed)


class ILoginPasswordExtractionPlugin_conformance:

    def test_ILoginPasswordExtractionPlugin_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import ILoginPasswordExtractionPlugin

        verifyClass( ILoginPasswordExtractionPlugin, self._getTargetClass() )

    def test_ILoginPasswordExtractionPlugin_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import ILoginPasswordExtractionPlugin

        listed = self._makeOne().listInterfaces()
        self.failUnless(ILoginPasswordExtractionPlugin.__name__ in listed)


class ILoginPasswordHostExtractionPlugin_conformance:

    def test_ILoginPasswordHostExtractionPlugin_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import ILoginPasswordHostExtractionPlugin

        verifyClass( ILoginPasswordHostExtractionPlugin
                   , self._getTargetClass() )

    def test_ILoginPasswordHostExtractionPlugin_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import ILoginPasswordHostExtractionPlugin

        listed = self._makeOne().listInterfaces()
        self.failUnless(ILoginPasswordHostExtractionPlugin.__name__ in listed)


class IChallengePlugin_conformance:

    def test_IChallengePlugin_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import IChallengePlugin

        verifyClass( IChallengePlugin, self._getTargetClass() )

    def test_IChallengePlugin_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import IChallengePlugin

        listed = self._makeOne().listInterfaces()
        self.failUnless(IChallengePlugin.__name__ in listed)


class ICredentialsUpdatePlugin_conformance:

    def test_ICredentialsUpdatePlugin_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import ICredentialsUpdatePlugin

        verifyClass( ICredentialsUpdatePlugin, self._getTargetClass() )

    def test_ICredentialsUpdatePlugin_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import ICredentialsUpdatePlugin

        listed = self._makeOne().listInterfaces()
        self.failUnless(ICredentialsUpdatePlugin.__name__ in listed)


class ICredentialsResetPlugin_conformance:

    def test_ICredentialsResetPlugin_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import ICredentialsResetPlugin

        verifyClass( ICredentialsResetPlugin, self._getTargetClass() )

    def test_ICredentialsResetPlugin_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import ICredentialsResetPlugin

        listed = self._makeOne().listInterfaces()
        self.failUnless(ICredentialsResetPlugin.__name__ in listed)


class IAuthenticationPlugin_conformance:

    def test_AuthenticationPlugin_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import IAuthenticationPlugin

        verifyClass( IAuthenticationPlugin, self._getTargetClass() )

    def test_IAuthenticationPlugin_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import IAuthenticationPlugin

        listed = self._makeOne().listInterfaces()
        self.failUnless(IAuthenticationPlugin.__name__ in listed)


class IUserEnumerationPlugin_conformance:

    def test_UserEnumerationPlugin_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import IUserEnumerationPlugin

        verifyClass( IUserEnumerationPlugin, self._getTargetClass() )

    def test_IUserEnumerationPlugin_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import IUserEnumerationPlugin

        listed = self._makeOne().listInterfaces()
        self.failUnless(IUserEnumerationPlugin.__name__ in listed)


class IUserAdderPlugin_conformance:

    def test_UserAdderPlugin_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import IUserAdderPlugin

        verifyClass( IUserAdderPlugin, self._getTargetClass() )

    def test_IUserAdderPlugin_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import IUserAdderPlugin

        listed = self._makeOne().listInterfaces()
        self.failUnless(IUserAdderPlugin.__name__ in listed)


class IGroupEnumerationPlugin_conformance:

    def test_GroupEnumerationPlugin_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import IGroupEnumerationPlugin

        verifyClass( IGroupEnumerationPlugin, self._getTargetClass() )

    def test_IGroupEnumerationPlugin_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import IGroupEnumerationPlugin

        listed = self._makeOne().listInterfaces()
        self.failUnless(IGroupEnumerationPlugin.__name__ in listed)


class IGroupsPlugin_conformance:

    def test_GroupsPlugin_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import IGroupsPlugin

        verifyClass( IGroupsPlugin, self._getTargetClass() )

    def test_IGroupsPlugin_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import IGroupsPlugin

        listed = self._makeOne().listInterfaces()
        self.failUnless(IGroupsPlugin.__name__ in listed)


class IRoleEnumerationPlugin_conformance:

    def test_RoleEnumerationPlugin_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import IRoleEnumerationPlugin

        verifyClass( IRoleEnumerationPlugin, self._getTargetClass() )

    def test_IRoleEnumerationPlugin_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import IRoleEnumerationPlugin

        listed = self._makeOne().listInterfaces()
        self.failUnless(IRoleEnumerationPlugin.__name__ in listed)


class IRolesPlugin_conformance:

    def test_RolesPlugin_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import IRolesPlugin

        verifyClass( IRolesPlugin, self._getTargetClass() )

    def test_IRolesPlugin_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import IRolesPlugin

        listed = self._makeOne().listInterfaces()
        self.failUnless(IRolesPlugin.__name__ in listed)


class IRoleAssignerPlugin_conformance:

    def test_RoleAssignerPlugin_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import IRoleAssignerPlugin

        verifyClass( IRoleAssignerPlugin, self._getTargetClass() )

    def test_IRoleAssignerPlugin_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import IRoleAssignerPlugin

        listed = self._makeOne().listInterfaces()
        self.failUnless(IRoleAssignerPlugin.__name__ in listed)


class IChallengeProtocolChooser_conformance:

    def test_ChallengeProtocolChooser_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import IChallengeProtocolChooser

        verifyClass( IChallengeProtocolChooser, self._getTargetClass() )

    def test_IChallengeProtocolChooser_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import IChallengeProtocolChooser

        listed = self._makeOne().listInterfaces()
        self.failUnless(IChallengeProtocolChooser.__name__ in listed)


class IRequestTypeSniffer_conformance:

    def test_RequestTypeSniffer_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import IRequestTypeSniffer

        verifyClass( IRequestTypeSniffer, self._getTargetClass() )

    def test_IRequestTypeSniffer_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import IRequestTypeSniffer

        listed = self._makeOne().listInterfaces()
        self.failUnless(IRequestTypeSniffer.__name__ in listed)


class IUserFolder_conformance:

    def test_conformance_IUserFolder( self ):

        from Products.PluggableAuthService.interfaces.authservice \
            import IUserFolder

        verifyClass( IUserFolder, self._getTargetClass() )


class IBasicUser_conformance:

    def test_conformance_IBasicUser( self ):

        from Products.PluggableAuthService.interfaces.authservice \
            import IBasicUser

        verifyClass( IBasicUser, self._getTargetClass() )


class IPropertiedUser_conformance:

    def test_conformance_IPropertiedUser( self ):

        from Products.PluggableAuthService.interfaces.authservice \
            import IPropertiedUser

        verifyClass( IPropertiedUser, self._getTargetClass() )


class IPropertySheet_conformance:

    def test_conformance_IPropertySheet( self ):

        from Products.PluggableAuthService.interfaces.propertysheets \
            import IPropertySheet

        verifyClass( IPropertySheet, self._getTargetClass() )


class INotCompetentPlugin_conformance:

    def test_INotCompetentPlugin_conformance( self ):

        from Products.PluggableAuthService.interfaces.plugins \
            import INotCompetentPlugin

        verifyClass( INotCompetentPlugin, self._getTargetClass() )

    def test_INotCompetentPlugin_listInterfaces(self):

        from Products.PluggableAuthService.interfaces.plugins \
            import INotCompetentPlugin

        listed = self._makeOne().listInterfaces()
        self.failUnless(INotCompetentPlugin.__name__ in listed)




##############################################################################
#
# Copyright (c) 2006 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from Testing import ZopeTestCase

ZopeTestCase.installProduct('PluginRegistry')
ZopeTestCase.installProduct('PluggableAuthService')
ZopeTestCase.installProduct('StandardCacheManagers')
ZopeTestCase.installProduct('GenericSetup')

from Testing.ZopeTestCase import user_name
from Testing.ZopeTestCase import user_password
from Testing.ZopeTestCase import user_role

from base64 import encodestring
user_auth = encodestring('%s:%s' % (user_name, user_password)).rstrip()

from Products.PluggableAuthService.interfaces.plugins import \
    IAuthenticationPlugin, IUserEnumerationPlugin, IRolesPlugin, \
    IRoleEnumerationPlugin, IRoleAssignerPlugin, \
    IChallengePlugin, IExtractionPlugin, IUserAdderPlugin


class PASTestCase(ZopeTestCase.ZopeTestCase):
    """ZopeTestCase with a PAS instead of the default user folder
    """

    def _setupUserFolder(self):
        """Creates a Pluggable Auth Service."""
        factory = self.folder.manage_addProduct['PluggableAuthService']
        factory.addPluggableAuthService()
        pas = self.folder.acl_users
        factory = pas.manage_addProduct['PluggableAuthService']
        factory.addHTTPBasicAuthHelper('http_auth')
        factory.addZODBUserManager('users')
        factory.addZODBRoleManager('roles')
        plugins = pas.plugins
        plugins.activatePlugin(IChallengePlugin, 'http_auth')
        plugins.activatePlugin(IExtractionPlugin, 'http_auth')
        plugins.activatePlugin(IUserAdderPlugin, 'users')
        plugins.activatePlugin(IAuthenticationPlugin, 'users')
        plugins.activatePlugin(IUserEnumerationPlugin, 'users')
        plugins.activatePlugin(IRolesPlugin, 'roles')
        plugins.activatePlugin(IRoleAssignerPlugin, 'roles')
        plugins.activatePlugin(IRoleEnumerationPlugin, 'roles')

    def _setupUser(self):
        """Creates the default user."""
        # OMFG, why doesn't PAS support userFolderAddUser?
        uf = self.folder.acl_users
        uf._doAddUser(user_name, user_password, [user_role], [])




##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import unittest


class ConformsToIPASEvent:

    def test_class_conforms_to_IPASEvent(self):
        from zope.interface.verify import verifyClass
        from Products.PluggableAuthService.interfaces.events import IPASEvent
        verifyClass(IPASEvent, self._getTargetClass())

    def test_instance_conforms_to_IPASEvent(self):
        from zope.interface.verify import verifyObject
        from Products.PluggableAuthService.interfaces.events import IPASEvent
        verifyObject(IPASEvent, self._makeOne())


class PASEventTests(unittest.TestCase, ConformsToIPASEvent):

    def _getTargetClass(self):
        from Products.PluggableAuthService.events import PASEvent
        return PASEvent

    def _makeOne(self, principal=None):
        if principal is None:
            principal = DummyPrincipal()
        return self._getTargetClass()(principal)


class PrincipalCreatedTests(unittest.TestCase, ConformsToIPASEvent):

    def _getTargetClass(self):
        from Products.PluggableAuthService.events import PrincipalCreated
        return PrincipalCreated

    def _makeOne(self, principal=None):
        if principal is None:
            principal = DummyPrincipal()
        return self._getTargetClass()(principal)

    def test_class_conforms_to_IPrincipalCreatedEvent(self):
        from zope.interface.verify import verifyClass
        from Products.PluggableAuthService.interfaces.events \
            import IPrincipalCreatedEvent
        verifyClass(IPrincipalCreatedEvent, self._getTargetClass())

    def test_instance_conforms_to_IPrincipalCreatedEvent(self):
        from zope.interface.verify import verifyObject
        from Products.PluggableAuthService.interfaces.events \
            import IPrincipalCreatedEvent
        verifyObject(IPrincipalCreatedEvent, self._makeOne())


class PrincipalDeletedTests(unittest.TestCase, ConformsToIPASEvent):

    def _getTargetClass(self):
        from Products.PluggableAuthService.events import PrincipalDeleted
        return PrincipalDeleted

    def _makeOne(self, principal=None):
        if principal is None:
            principal = DummyPrincipal()
        return self._getTargetClass()(principal)

    def test_class_conforms_to_IPrincipalDeletedEvent(self):
        from zope.interface.verify import verifyClass
        from Products.PluggableAuthService.interfaces.events \
            import IPrincipalDeletedEvent
        verifyClass(IPrincipalDeletedEvent, self._getTargetClass())

    def test_instance_conforms_to_IPrincipalDeletedEvent(self):
        from zope.interface.verify import verifyObject
        from Products.PluggableAuthService.interfaces.events \
            import IPrincipalDeletedEvent
        verifyObject(IPrincipalDeletedEvent, self._makeOne())


class CredentialsUpdatedTests(unittest.TestCase, ConformsToIPASEvent):

    def _getTargetClass(self):
        from Products.PluggableAuthService.events import CredentialsUpdated
        return CredentialsUpdated

    def _makeOne(self, principal=None, password='password'):
        if principal is None:
            principal = DummyPrincipal()
        return self._getTargetClass()(principal, password)

    def test_class_conforms_to_ICredentialsUpdatedEvent(self):
        from zope.interface.verify import verifyClass
        from Products.PluggableAuthService.interfaces.events \
            import ICredentialsUpdatedEvent
        verifyClass(ICredentialsUpdatedEvent, self._getTargetClass())

    def test_instance_conforms_to_ICredentialsUpdatedEvent(self):
        from zope.interface.verify import verifyObject
        from Products.PluggableAuthService.interfaces.events \
            import ICredentialsUpdatedEvent
        verifyObject(ICredentialsUpdatedEvent, self._makeOne())


class PropertiesUpdatedTests(unittest.TestCase, ConformsToIPASEvent):

    def _getTargetClass(self):
        from Products.PluggableAuthService.events import PropertiesUpdated
        return PropertiesUpdated

    def _makeOne(self, principal=None, properties=None):
        if principal is None:
            principal = DummyPrincipal()
        if properties is None:
            properties = {}
        return self._getTargetClass()(principal, properties)

    def test_class_conforms_to_IPropertiesUpdatedEvent(self):
        from zope.interface.verify import verifyClass
        from Products.PluggableAuthService.interfaces.events \
            import IPropertiesUpdatedEvent
        verifyClass(IPropertiesUpdatedEvent, self._getTargetClass())

    def test_instance_conforms_to_IPropertiesUpdatedEvent(self):
        from zope.interface.verify import verifyObject
        from Products.PluggableAuthService.interfaces.events \
            import IPropertiesUpdatedEvent
        verifyObject(IPropertiesUpdatedEvent, self._makeOne())


class DummyPrincipal(object):
    pass


class GroupDeletedTests(unittest.TestCase, ConformsToIPASEvent):

    def _getTargetClass(self):
        from Products.PluggableAuthService.events import GroupDeleted
        return GroupDeleted

    def _makeOne(self, Group=None):
        if Group is None:
            Group = DummyGroup()
        return self._getTargetClass()(Group)

    def test_class_conforms_to_IGroupDeletedEvent(self):
        from zope.interface.verify import verifyClass
        from Products.PluggableAuthService.interfaces.events \
            import IGroupDeletedEvent
        verifyClass(IGroupDeletedEvent, self._getTargetClass())

    def test_instance_conforms_to_IGroupDeletedEvent(self):
        from zope.interface.verify import verifyObject
        from Products.PluggableAuthService.interfaces.events \
            import IGroupDeletedEvent
        verifyObject(IGroupDeletedEvent, self._makeOne())


class DummyGroup(object):
    pass




##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import unittest

from Acquisition import Implicit, aq_base, aq_parent
from OFS.ObjectManager import ObjectManager

from Products.PluggableAuthService.tests.conformance \
    import IRolesPlugin_conformance

class FauxObject( Implicit ):

    def __init__( self, id=None ):

        self._id = id

    def getId( self ):

        return self._id

    def __repr__( self ):

        return '<FauxObject: %s>' % self._id

    def publishable( self, *args, **kw ):

        return 'Args: %s\nKeywords: %s' % ( args, kw )

class FauxContainer( FauxObject, ObjectManager ):

    pass

class FauxRoot( FauxContainer ):

    isTopLevelPrincipiaApplicationObject = 1

    def getPhysicalRoot( self ):
        return self

    def getPhysicalPath( self ):
        return ()

class FauxUser( Implicit ):

    def __init__( self, id, name=None, roles={} ):

        self._id = id
        self._name = name
        self._roles = roles

    def getId( self ):

        return self._id

    def getUserName( self ):

        return self._name

    def getRoles( self ):

        return self._roles.keys()

    def _addRoles(self, roles):
        for role in roles:
            self._roles[role] = 1

    def __repr__( self ):

        return '<FauxUser: %s>' % self._id

class LocalRolePluginTestCase( unittest.TestCase
                             , IRolesPlugin_conformance
                             ):

    def _getTargetClass( self ):

        from Products.PluggableAuthService.plugins.LocalRolePlugin \
             import LocalRolePlugin

        return LocalRolePlugin

    def _makeOne( self, id='test', *args, **kw ):

        return self._getTargetClass()( id, *args, **kw )

    def _makeTree( self ):

        rc = FauxObject( 'rc' )
        root = FauxRoot( 'root' ).__of__( rc )
        folder = FauxContainer( 'folder' ).__of__( root )
        object = FauxObject( 'object' ).__of__( folder )

        return rc, root, folder, object

    def test_no_local_roles( self ):

        rc, root, folder, object = self._makeTree()

        lrp = self._makeOne( 'no_roles' ).__of__(root)

        user = FauxUser( 'loser' ).__of__(root)

        self.assertEquals( lrp.getRolesForPrincipal( user ), None )

    def test_some_local_roles( self ):

        rc, root, folder, object = self._makeTree()

        root.__ac_local_roles__ = { 'some_manager' : [ 'Manager' ] }

        lrp = self._makeOne( 'roles' ).__of__(root)

        user = FauxUser( 'some_manager' ).__of__(root)

        self.assertEquals( lrp.getRolesForPrincipal( user ), [ 'Manager' ] )

if __name__ == "__main__":
    unittest.main()


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite( LocalRolePluginTestCase ),
        ))



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Classes: ZODBUserManager

$Id$
"""
import copy
import logging
try:
    from hashlib import sha1 as sha
except:
    from sha import sha

from AccessControl import ClassSecurityInfo, AuthEncoding
from AccessControl.requestmethod import postonly
from AccessControl.SecurityManagement import getSecurityManager
from App.class_init import default__class_init__ as InitializeClass
from BTrees.OOBTree import OOBTree
from OFS.Cache import Cacheable

from zope.interface import Interface

from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from Products.PluggableAuthService.interfaces.plugins \
    import IAuthenticationPlugin
from Products.PluggableAuthService.interfaces.plugins \
    import IUserEnumerationPlugin
from Products.PluggableAuthService.interfaces.plugins \
    import IUserAdderPlugin

from Products.PluggableAuthService.permissions import ManageUsers
from Products.PluggableAuthService.permissions import SetOwnPassword
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.utils import classImplements
from Products.PluggableAuthService.utils import createViewName
from Products.PluggableAuthService.utils import csrf_only

logger = logging.getLogger('PluggableAuthService')


class IZODBUserManager(Interface):
    """ Marker interface.
    """


manage_addZODBUserManagerForm = PageTemplateFile(
    'www/zuAdd', globals(), __name__='manage_addZODBUserManagerForm' )

def addZODBUserManager( dispatcher, id, title=None, REQUEST=None ):
    """ Add a ZODBUserManagern to a Pluggable Auth Service. """

    zum = ZODBUserManager(id, title)
    dispatcher._setObject(zum.getId(), zum)

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(
                                '%s/manage_workspace'
                                '?manage_tabs_message='
                                'ZODBUserManager+added.'
                            % dispatcher.absolute_url())

class ZODBUserManager( BasePlugin, Cacheable ):

    """ PAS plugin for managing users in the ZODB.
    """

    meta_type = 'ZODB User Manager'

    security = ClassSecurityInfo()

    def __init__(self, id, title=None):

        self._id = self.id = id
        self.title = title

        self._user_passwords = OOBTree()
        self._login_to_userid = OOBTree()
        self._userid_to_login = OOBTree()

    #
    #   IAuthenticationPlugin implementation
    #
    security.declarePrivate( 'authenticateCredentials' )
    def authenticateCredentials( self, credentials ):

        """ See IAuthenticationPlugin.

        o We expect the credentials to be those returned by
          ILoginPasswordExtractionPlugin.
        """
        login = credentials.get( 'login' )
        password = credentials.get( 'password' )

        if login is None or password is None:
            return None

        # Do we have a link between login and userid?  Do NOT fall
        # back to using the login as userid when there is no match, as
        # that gives a high chance of seeming to log in successfully,
        # but in reality failing.
        userid = self._login_to_userid.get(login)
        if userid is None:
            # Someone may be logging in with a userid instead of a
            # login name and the two are not the same.  We could try
            # turning those around, but really we should just fail.
            #
            # userid = login
            # login = self._userid_to_login.get(userid)
            # if login is None:
            #     return None
            return None

        reference = self._user_passwords.get(userid)

        if reference is None:
            return None

        if AuthEncoding.is_encrypted( reference ):
            if AuthEncoding.pw_validate( reference, password ):
                return userid, login

        # Support previous naive behavior
        digested = sha( password ).hexdigest()

        if reference == digested:
            return userid, login

        return None

    #
    #   IUserEnumerationPlugin implementation
    #
    security.declarePrivate( 'enumerateUsers' )
    def enumerateUsers( self
                      , id=None
                      , login=None
                      , exact_match=False
                      , sort_by=None
                      , max_results=None
                      , **kw
                      ):

        """ See IUserEnumerationPlugin.
        """
        user_info = []
        user_ids = []
        plugin_id = self.getId()
        view_name = createViewName('enumerateUsers', id or login)


        if isinstance( id, basestring ):
            id = [ id ]

        if isinstance( login, basestring ):
            login = [ login ]

        # Look in the cache first...
        keywords = copy.deepcopy(kw)
        keywords.update( { 'id' : id
                         , 'login' : login
                         , 'exact_match' : exact_match
                         , 'sort_by' : sort_by
                         , 'max_results' : max_results
                         }
                       )
        cached_info = self.ZCacheable_get( view_name=view_name
                                         , keywords=keywords
                                         , default=None
                                         )
        if cached_info is not None:
            return tuple(cached_info)

        terms = id or login

        if exact_match:
            if terms:

                if id:
                    # if we're doing an exact match based on id, it
                    # absolutely will have been qualified (if we have a
                    # prefix), so we can ignore any that don't begin with
                    # our prefix
                    id = [ x for x in id if x.startswith(self.prefix) ]
                    user_ids.extend( [ x[len(self.prefix):] for x in id ] )
                elif login:
                    user_ids.extend( [ self._login_to_userid.get( x )
                                       for x in login ] )

                # we're claiming an exact match search, if we still don't
                # have anything, better bail.
                if not user_ids:
                    return ()
            else:
                # insane - exact match with neither login nor id
                return ()

        if user_ids:
            user_filter = None

        else:   # Searching
            user_ids = self.listUserIds()
            user_filter = _ZODBUserFilter( id, login, **kw )

        for user_id in user_ids:

            if self._userid_to_login.get( user_id ):
                e_url = '%s/manage_users' % self.getId()
                qs = 'user_id=%s' % user_id

                info = { 'id' : self.prefix + user_id
                       , 'login' : self._userid_to_login[ user_id ]
                       , 'pluginid' : plugin_id
                       , 'editurl' : '%s?%s' % (e_url, qs)
                       }

                if not user_filter or user_filter( info ):
                    user_info.append( info )

        # Put the computed value into the cache
        self.ZCacheable_set(user_info, view_name=view_name, keywords=keywords)

        return tuple( user_info )

    #
    #   IUserAdderPlugin implementation
    #
    security.declarePrivate( 'doAddUser' )
    def doAddUser( self, login, password ):
        try:
            self.addUser( login, login, password )
        except KeyError:
            return False
        return True

    #
    #   (notional)IZODBUserManager interface
    #
    security.declareProtected( ManageUsers, 'listUserIds' )
    def listUserIds( self ):

        """ -> ( user_id_1, ... user_id_n )
        """
        return self._user_passwords.keys()

    security.declareProtected( ManageUsers, 'getUserInfo' )
    def getUserInfo( self, user_id ):

        """ user_id -> {}
        """
        return { 'user_id' : user_id
               , 'login_name' : self._userid_to_login[ user_id ]
               , 'pluginid' : self.getId()
               }

    security.declareProtected( ManageUsers, 'listUserInfo' )
    def listUserInfo( self ):

        """ -> ( {}, ...{} )

        o Return one mapping per user, with the following keys:

          - 'user_id'
          - 'login_name'
        """
        return [ self.getUserInfo( x ) for x in self._user_passwords.keys() ]

    security.declareProtected( ManageUsers, 'getUserIdForLogin' )
    def getUserIdForLogin( self, login_name ):

        """ login_name -> user_id

        o Raise KeyError if no user exists for the login name.
        """
        return self._login_to_userid[ login_name ]

    security.declareProtected( ManageUsers, 'getLoginForUserId' )
    def getLoginForUserId( self, user_id ):

        """ user_id -> login_name

        o Raise KeyError if no user exists for that ID.
        """
        return self._userid_to_login[ user_id ]

    security.declarePrivate( 'addUser' )
    def addUser( self, user_id, login_name, password ):

        if self._user_passwords.get( user_id ) is not None:
            raise KeyError, 'Duplicate user ID: %s' % user_id

        if self._login_to_userid.get( login_name ) is not None:
            raise KeyError, 'Duplicate login name: %s' % login_name

        self._user_passwords[ user_id ] = self._pw_encrypt( password)
        self._login_to_userid[ login_name ] = user_id
        self._userid_to_login[ user_id ] = login_name

        # enumerateUsers return value has changed
        view_name = createViewName('enumerateUsers')
        self.ZCacheable_invalidate(view_name=view_name)

    security.declarePrivate('updateUser')
    def updateUser(self, user_id, login_name):

        # The following raises a KeyError if the user_id is invalid
        old_login = self.getLoginForUserId(user_id)

        if old_login != login_name:

            if self._login_to_userid.get(login_name) is not None:
                raise ValueError('Login name not available: %s' % login_name)

            del self._login_to_userid[old_login]
            self._login_to_userid[login_name] = user_id
            self._userid_to_login[user_id] = login_name
        # Signal success.
        return True

    security.declarePrivate('updateEveryLoginName')
    def updateEveryLoginName(self, quit_on_first_error=True):
        # Update all login names to their canonical value.  This
        # should be done after changing the login_transform property
        # of pas.  You can set quit_on_first_error to False to report
        # all errors before quitting with an error.  This can be
        # useful if you want to know how many problems there are, if
        # any.
        pas = self._getPAS()
        transform = pas._get_login_transform_method()
        if not transform:
            logger.warn("PAS has a non-existing, empty or wrong "
                        "login_transform property.")
            return

        # Make a fresh mapping, as we do not want to add or remove
        # items to the original mapping while we are iterating over
        # it.
        new_login_to_userid = OOBTree()
        errors = []
        for old_login_name, user_id in self._login_to_userid.items():
            new_login_name = transform(old_login_name)
            if new_login_name in new_login_to_userid:
                logger.error("User id %s: login name %r already taken.",
                             user_id, new_login_name)
                errors.append(new_login_name)
                if quit_on_first_error:
                    break
            new_login_to_userid[new_login_name] = user_id
            if new_login_name != old_login_name:
                self._userid_to_login[user_id] = new_login_name
                # Also, remove from the cache
                view_name = createViewName('enumerateUsers', user_id)
                self.ZCacheable_invalidate(view_name=view_name)
                logger.debug("User id %s: changed login name from %r to %r.",
                             user_id, old_login_name, new_login_name)

        # If there were errors, we do not want to save any changes.
        if errors:
            logger.error("There were %d errors when updating login names. "
                         "quit_on_first_error was %r", len(errors),
                         quit_on_first_error)
            # Make sure the exception we raise is not swallowed.
            self._dont_swallow_my_exceptions = True
            raise ValueError("Transformed login names are not unique: %s." %
                             ', '.join(errors))

        # Make sure we did not lose any users.
        assert(len(self._login_to_userid.keys())
               == len(new_login_to_userid.keys()))
        # Empty the main cache.
        view_name = createViewName('enumerateUsers')
        self.ZCacheable_invalidate(view_name=view_name)
        # Store the new login mapping.
        self._login_to_userid = new_login_to_userid

    security.declarePrivate( 'removeUser' )
    def removeUser( self, user_id ):

        if self._user_passwords.get( user_id ) is None:
            raise KeyError, 'Invalid user ID: %s' % user_id

        login_name = self._userid_to_login[ user_id ]

        del self._user_passwords[ user_id ]
        del self._login_to_userid[ login_name ]
        del self._userid_to_login[ user_id ]

        # Also, remove from the cache
        view_name = createViewName('enumerateUsers')
        self.ZCacheable_invalidate(view_name=view_name)
        view_name = createViewName('enumerateUsers', user_id)
        self.ZCacheable_invalidate(view_name=view_name)

    security.declarePrivate( 'updateUserPassword' )
    def updateUserPassword( self, user_id, password ):

        if self._user_passwords.get( user_id ) is None:
            raise KeyError, 'Invalid user ID: %s' % user_id

        if password:
            self._user_passwords[ user_id ] = self._pw_encrypt( password )

    security.declarePrivate( '_pw_encrypt' )
    def _pw_encrypt( self, password ):
        """Returns the AuthEncoding encrypted password

        If 'password' is already encrypted, it is returned
        as is and not encrypted again.
        """
        if AuthEncoding.is_encrypted( password ):
            return password
        return AuthEncoding.pw_encrypt( password )

    #
    #   ZMI
    #
    manage_options = ( ( { 'label': 'Users',
                           'action': 'manage_users', }
                         ,
                       )
                     + BasePlugin.manage_options
                     + Cacheable.manage_options
                     )

    security.declarePublic( 'manage_widgets' )
    manage_widgets = PageTemplateFile( 'www/zuWidgets'
                                     , globals()
                                     , __name__='manage_widgets'
                                     )

    security.declareProtected( ManageUsers, 'manage_users' )
    manage_users = PageTemplateFile( 'www/zuUsers'
                                   , globals()
                                   , __name__='manage_users'
                                   )

    security.declareProtected( ManageUsers, 'manage_addUser' )
    @csrf_only
    @postonly
    def manage_addUser( self
                      , user_id
                      , login_name
                      , password
                      , confirm
                      , RESPONSE=None
                      , REQUEST=None
                      ):
        """ Add a user via the ZMI.
        """
        if password != confirm:
            message = 'password+and+confirm+do+not+match'

        else:

            if not login_name:
                login_name = user_id

            # XXX:  validate 'user_id', 'login_name' against policies?

            self.addUser( user_id, login_name, password )

            message = 'User+added'

        if RESPONSE is not None:
            RESPONSE.redirect( '%s/manage_users?manage_tabs_message=%s'
                             % ( self.absolute_url(), message )
                             )

    security.declareProtected( ManageUsers, 'manage_updateUserPassword' )
    @csrf_only
    @postonly
    def manage_updateUserPassword( self
                                 , user_id
                                 , password
                                 , confirm
                                 , RESPONSE=None
                                 , REQUEST=None
                                 ):
        """ Update a user's login name / password via the ZMI.
        """
        if password and password != confirm:
            message = 'password+and+confirm+do+not+match'

        else:

            self.updateUserPassword( user_id, password )

            message = 'password+updated'

        if RESPONSE is not None:
            RESPONSE.redirect( '%s/manage_users?manage_tabs_message=%s'
                             % ( self.absolute_url(), message )
                             )

    security.declareProtected( ManageUsers, 'manage_updateUser' )
    @csrf_only
    @postonly
    def manage_updateUser(self
                         , user_id
                         , login_name
                         , RESPONSE=None
                         , REQUEST=None
                         ):
        """ Update a user's login name via the ZMI.
        """
        if not login_name:
            login_name = user_id

        # XXX:  validate 'user_id', 'login_name' against policies?

        self.updateUser(user_id, login_name)

        message = 'Login+name+updated'

        if RESPONSE is not None:
            RESPONSE.redirect( '%s/manage_users?manage_tabs_message=%s'
                             % ( self.absolute_url(), message )
                             )

    security.declareProtected( ManageUsers, 'manage_removeUsers' )
    @csrf_only
    @postonly
    def manage_removeUsers( self
                          , user_ids
                          , RESPONSE=None
                          , REQUEST=None
                          ):
        """ Remove one or more users via the ZMI.
        """
        user_ids = filter( None, user_ids )

        if not user_ids:
            message = 'no+users+selected'

        else:

            for user_id in user_ids:
                self.removeUser( user_id )

            message = 'Users+removed'

        if RESPONSE is not None:
            RESPONSE.redirect( '%s/manage_users?manage_tabs_message=%s'
                             % ( self.absolute_url(), message )
                             )

    #
    #   Allow users to change their own login name and password.
    #
    security.declareProtected( SetOwnPassword, 'getOwnUserInfo' )
    def getOwnUserInfo( self ):

        """ Return current user's info.
        """
        user_id = getSecurityManager().getUser().getId()

        return self.getUserInfo( user_id )

    security.declareProtected( SetOwnPassword, 'manage_updatePasswordForm' )
    manage_updatePasswordForm = PageTemplateFile( 'www/zuPasswd'
                                   , globals()
                                   , __name__='manage_updatePasswordForm'
                                   )

    security.declareProtected( SetOwnPassword, 'manage_updatePassword' )
    @csrf_only
    @postonly
    def manage_updatePassword( self
                             , login_name
                             , password
                             , confirm
                             , RESPONSE=None
                             , REQUEST=None
                             ):
        """ Update the current user's password and login name.
        """
        user_id = getSecurityManager().getUser().getId()
        if password != confirm:
            message = 'password+and+confirm+do+not+match'

        else:

            if not login_name:
                login_name = user_id

            # XXX:  validate 'user_id', 'login_name' against policies?
            self.updateUser( user_id, login_name )
            self.updateUserPassword( user_id, password )

            message = 'password+updated'

        if RESPONSE is not None:
            RESPONSE.redirect( '%s/manage_updatePasswordForm'
                               '?manage_tabs_message=%s'
                             % ( self.absolute_url(), message )
                             )

classImplements( ZODBUserManager
               , IZODBUserManager
               , IAuthenticationPlugin
               , IUserEnumerationPlugin
               , IUserAdderPlugin
               )

InitializeClass( ZODBUserManager )

class _ZODBUserFilter:

    def __init__( self
                , id=None
                , login=None
                , **kw
                ):

        self._filter_ids = id
        self._filter_logins = login
        self._filter_keywords = kw

    def __call__( self, user_info ):

        if self._filter_ids:

            key = 'id'
            to_test = self._filter_ids

        elif self._filter_logins:

            key = 'login'
            to_test = self._filter_logins

        elif self._filter_keywords:
            return 0    # TODO:  try using 'kw'

        else:
            return 1    # the search is done without any criteria

        value = user_info.get( key )

        if not value:
            return 0

        for contained in to_test:
            if value.lower().find( contained.lower() ) >= 0:
                return 1

        return 0



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import unittest

from OFS.SimpleItem import SimpleItem
from Products.PluggableAuthService.tests.conformance \
    import IGroupsPlugin_conformance

from Products.PluggableAuthService.tests.test_PluggableAuthService \
    import FauxContainer

class FauxPAS( FauxContainer ):

    def __init__( self ):
        self._id = 'acl_users'

    def _getGroupsForPrincipal( self, principal, request=None, plugins=None
                              , ignore_plugins=None ):
        all_groups = []

        if ignore_plugins is None:
            ignore_plugins = ()

        for plugin in self.objectValues():
            if plugin.getId() in ignore_plugins:
                continue
            groups = plugin.getGroupsForPrincipal( principal, request )
            groups = [ '%s:%s' % ( plugin.getId(), x ) for x in groups ]
            all_groups.extend( groups )

        return all_groups

class DummyPrincipal:

    def __init__( self, id, groups=() ):
        self._id = id
        self._groups = groups

    def getId( self ):
        return self._id

    def getGroups( self ):
        return self._groups

class DummyGroupsManager( SimpleItem ):

    def __init__( self, id ):
        self._id = id
        self._groups = {}
        self._prin_to_group = {}

    def getId( self ):
        return self._id

    def addGroup( self, group_id ):
        self._groups[group_id] = 1

    def addPrincipalToGroup( self, prin, group_id ):
        prin_id = prin.getId()
        already_in = self._prin_to_group.get( prin_id, () )
        if group_id not in already_in:
            self._prin_to_group[prin_id] = already_in + ( group_id, )

    def getGroupsForPrincipal( self, prin, request=None ):
        prin_id = prin.getId()
        return self._prin_to_group.get( prin_id, () )

class RecursiveGroupsPluginTests( unittest.TestCase
                                , IGroupsPlugin_conformance
                                ):

    def _getTargetClass( self ):

        from Products.PluggableAuthService.plugins.RecursiveGroupsPlugin \
            import RecursiveGroupsPlugin

        return RecursiveGroupsPlugin

    def _makeOne( self, id='test', *args, **kw ):

        return self._getTargetClass()( id=id, *args, **kw )

    def test_simple_flattening( self ):

        pas = FauxPAS()
        dgm = DummyGroupsManager( 'dummy' )
        rgp = self._makeOne( 'rgp' )

        pas._setObject( 'dgm', dgm )
        pas._setObject( 'rgp', rgp )

        dgm = pas._getOb( 'dgm' )
        dgm.addGroup( 'group1' )
        dgm.addGroup( 'group2' )

        rgp = pas._getOb( 'rgp' )

        user = DummyPrincipal( 'user1', ( 'dummy:group1', ) )

        dgm.addPrincipalToGroup( user, 'group1' )
        dgm.addPrincipalToGroup( DummyPrincipal( 'dummy:group1' ), 'group2' )

        self.assertEqual( rgp.getGroupsForPrincipal( user )
                        , ( 'dummy:group1', 'dummy:group2' )
                        )

    def test_complex_flattening( self ):

        pas = FauxPAS()
        dgm = DummyGroupsManager( 'dummy' )
        odgm = DummyGroupsManager( 'other_dummy' )
        rgp = self._makeOne( 'rgp' )

        pas._setObject( 'dgm', dgm )
        pas._setObject( 'odgm', odgm )
        pas._setObject( 'rgp', rgp )

        dgm = pas._getOb( 'dgm' )
        dgm.addGroup( 'group1' )
        dgm.addGroup( 'group2' )

        odgm = pas._getOb( 'odgm' )
        odgm.addGroup( 'group3' )
        odgm.addGroup( 'group4' )

        rgp = pas._getOb( 'rgp' )

        user = DummyPrincipal( 'user1', ( 'dummy:group1'
                                        , 'other_dummy:group3' ) )

        dgm.addPrincipalToGroup( user, 'group1' )
        dgm.addPrincipalToGroup( DummyPrincipal( 'dummy:group1' ), 'group2' )

        odgm.addPrincipalToGroup( user, 'group3' )
        odgm.addPrincipalToGroup( DummyPrincipal( 'dummy:group2' ), 'group4' )

        groups = rgp.getGroupsForPrincipal( user )
        self.assertEqual( len( groups ), 4 )
        self.failUnless( 'dummy:group1' in groups )
        self.failUnless( 'dummy:group2' in groups )
        self.failUnless( 'other_dummy:group3' in groups )
        self.failUnless( 'other_dummy:group4' in groups )

    def test_cross_nested_flattening( self ):

        pas = FauxPAS()
        dgm = DummyGroupsManager( 'dummy' )
        odgm = DummyGroupsManager( 'other_dummy' )
        rgp = self._makeOne( 'rgp' )

        pas._setObject( 'dgm', dgm )
        pas._setObject( 'odgm', odgm )
        pas._setObject( 'rgp', rgp )

        dgm = pas._getOb( 'dgm' )
        dgm.addGroup( 'group1' )

        odgm = pas._getOb( 'odgm' )
        odgm.addGroup( 'group2' )

        rgp = pas._getOb( 'rgp' )

        user = DummyPrincipal( 'user1', ( 'dummy:group1'
                                        , 'other_dummy:group2' ) )

        dgm.addPrincipalToGroup( user, 'group1' )
        dgm.addPrincipalToGroup( DummyPrincipal( 'other_dummy:group1' )
                               , 'group1' )

        odgm.addPrincipalToGroup( user, 'group2' )
        odgm.addPrincipalToGroup( DummyPrincipal( 'dummy:group1' ), 'group2' )

        groups = rgp.getGroupsForPrincipal( user )
        self.assertEqual( len( groups ), 2 )
        self.failUnless( 'dummy:group1' in groups )
        self.failUnless( 'other_dummy:group2' in groups )

if __name__ == "__main__":
    unittest.main()


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite( RecursiveGroupsPluginTests ),
        ))



##############################################################################
#
# Copyright (c) 2001-2008 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Unit tests for DynamicGroupsPlugin

$Id$
"""
import unittest

from Products.PluggableAuthService.tests.conformance \
    import IGroupsPlugin_conformance
from Products.PluggableAuthService.tests.conformance \
    import IGroupEnumerationPlugin_conformance
from Products.PluggableAuthService.tests.utils import _setUpDefaultTraversable

from Products.PluggableAuthService.plugins.tests.helpers \
     import makeRequestAndResponse

class FauxScript:


    def __init__( self, id, return_value=0 ):

        self._id = id
        self.return_value = return_value

    def __call__( self, *args, **kw ):

        return self.return_value

class FauxPrincipal:

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__( self, id ):
        self._id = id

    def getId( self ):
        return self._id

class DynamicGroupsPlugin( unittest.TestCase
                         , IGroupsPlugin_conformance
                         , IGroupEnumerationPlugin_conformance
                         ):

    def _getTargetClass( self ):

        from Products.PluggableAuthService.plugins.DynamicGroupsPlugin \
            import DynamicGroupsPlugin

        return DynamicGroupsPlugin

    def _makeOne( self, id='test', *args, **kw ):

        return self._getTargetClass()( id, *args, **kw )

    def test_empty( self ):

        dpg = self._makeOne( 'empty' )

        self.assertEqual( len( dpg.listGroupIds() ), 0 )
        self.assertEqual( len( dpg.listGroupInfo() ), 0 )
        self.assertEqual( len( dpg.enumerateGroups() ), 0 )

    def test_addGroup_simple( self ):

        dpg = self._makeOne( 'adding' )

        dpg.addGroup( 'everyone', 'python:True', 'title', 'description', True )

        self.assertEqual( len( dpg.listGroupIds() ), 1 )
        self.failUnless( 'everyone' in dpg.listGroupIds() )

        self.assertEqual( len( dpg.listGroupInfo() ), 1 )
        info = dpg.listGroupInfo()[0]

        self.assertEqual( info[ 'id' ], 'everyone' )
        self.assertEqual( info[ 'title' ], 'title' )
        self.assertEqual( info[ 'description' ], 'description' )
        self.assertEqual( info[ 'predicate' ], 'python:True' )
        self.assertEqual( info[ 'active' ], True )

    def test_addGroup_duplicate( self ):

        dpg = self._makeOne( 'adding_duplicate' )

        dpg.addGroup( 'everyone', 'python:True', 'title', 'description', True )

        self.assertRaises( KeyError
                         , dpg.addGroup
                         , 'everyone'
                         , 'python:False'
                         , 'other title'
                         , 'other descripton'
                         , False )

        self.assertEqual( len( dpg.listGroupInfo() ), 1 )
        info = dpg.listGroupInfo()[0]

        self.assertEqual( info[ 'id' ], 'everyone' )
        self.assertEqual( info[ 'predicate' ], 'python:True' )
        self.assertEqual( info[ 'title' ], 'title' )
        self.assertEqual( info[ 'description' ], 'description' )
        self.assertEqual( info[ 'active' ], True )

    def test_removeGroup_nonesuch( self ):

        dpg = self._makeOne( 'removing_nonesuch' )

        self.assertRaises( KeyError, dpg.removeGroup, 'everyone' )

    def test_removeGroup( self ):

        dpg = self._makeOne( 'removing' )

        dpg.addGroup( 'everyone', predicate='python:True' )
        dpg.addGroup( 'beast', predicate='python:666' )
        dpg.addGroup( 'noone', predicate='python:False' )

        self.assertEqual( len( dpg.listGroupIds() ), 3 )
        self.failUnless( 'everyone' in dpg.listGroupIds() )
        self.failUnless( 'beast' in dpg.listGroupIds() )
        self.failUnless( 'noone' in dpg.listGroupIds() )

        info_list = dpg.listGroupInfo()
        self.assertEqual( len( info_list ), 3 )

        ids = [ x[ 'id' ] for x in info_list ]
        self.failUnless( 'everyone' in ids )
        self.failUnless( 'beast' in ids )
        self.failUnless( 'noone' in ids )

        dpg.removeGroup( 'beast' )

        self.assertEqual( len( dpg.listGroupIds() ), 2 )
        self.failUnless( 'everyone' in dpg.listGroupIds() )
        self.failIf( 'beast' in dpg.listGroupIds() )
        self.failUnless( 'noone' in dpg.listGroupIds() )

        info_list = dpg.listGroupInfo()
        self.assertEqual( len( info_list ), 2 )

        ids = [ x[ 'id' ] for x in info_list ]
        self.failUnless( 'everyone' in ids )
        self.failIf( 'beast' in ids )
        self.failUnless( 'noone' in ids )

    def test_updateGroup_nonesuch( self ):

        dpg = self._makeOne( 'updating_nonesuch' )

        self.assertRaises( KeyError, dpg.updateGroup
                         , 'everyone', 'title', True )

    def test_updateGroup_simple( self ):

        dpg = self._makeOne( 'updating_simple' )

        dpg.addGroup( 'noone', predicate='python:False' )

        self.assertEqual( len( dpg.listGroupIds() ), 1 )
        self.failUnless( 'noone' in dpg.listGroupIds() )

        self.assertEqual( len( dpg.listGroupInfo() ), 1 )
        info = dpg.listGroupInfo()[0]

        self.assertEqual( info[ 'id' ], 'noone' )
        self.assertEqual( info[ 'predicate' ], 'python:False' )
        self.assertEqual( info[ 'title' ], '' )
        self.assertEqual( info[ 'description' ], '' )
        self.assertEqual( info[ 'active' ], True )

        dpg.updateGroup( 'noone', 'python:True', 'title', 'description', False )

        self.assertEqual( len( dpg.listGroupIds() ), 1 )
        self.failUnless( 'noone' in dpg.listGroupIds() )

        self.assertEqual( len( dpg.listGroupInfo() ), 1 )
        info = dpg.listGroupInfo()[0]

        self.assertEqual( info[ 'id' ], 'noone' )
        self.assertEqual( info[ 'predicate' ], 'python:True' )
        self.assertEqual( info[ 'title' ], 'title' )
        self.assertEqual( info[ 'description' ], 'description' )
        self.assertEqual( info[ 'active' ], False )

    def test_updateGroup_partial( self ):

        dpg = self._makeOne( 'updating_partial' )

        dpg.addGroup( 'everyone', 'python:True', 'title', 'description', True )

        self.assertEqual( len( dpg.listGroupIds() ), 1 )
        self.failUnless( 'everyone' in dpg.listGroupIds() )

        self.assertEqual( len( dpg.listGroupInfo() ), 1 )
        info = dpg.listGroupInfo()[0]

        self.assertEqual( info[ 'id' ], 'everyone' )
        self.assertEqual( info[ 'predicate' ], 'python:True' )
        self.assertEqual( info[ 'title' ], 'title' )
        self.assertEqual( info[ 'description' ], 'description' )
        self.assertEqual( info[ 'active' ], True )

        dpg.updateGroup( 'everyone', predicate='python:False' )

        self.assertEqual( len( dpg.listGroupIds() ), 1 )
        self.failUnless( 'everyone' in dpg.listGroupIds() )

        self.assertEqual( len( dpg.listGroupInfo() ), 1 )
        info = dpg.listGroupInfo()[0]

        self.assertEqual( info[ 'id' ], 'everyone' )
        self.assertEqual( info[ 'predicate' ], 'python:False' )
        self.assertEqual( info[ 'title' ], 'title' )
        self.assertEqual( info[ 'description' ], 'description' )
        self.assertEqual( info[ 'active' ], True )

    def test_enumerateGroups_all( self ):

        from Products.PluggableAuthService.tests.test_PluggableAuthService \
            import FauxRoot

        root = FauxRoot()
        dpg = self._makeOne( 'enumerating' ).__of__( root )

        dpg.addGroup( 'everyone', 'python:True', 'Everyone', '', True )
        dpg.addGroup( 'noone', 'python:False', active=True )
        dpg.addGroup( 'hohum', 'nothing', active=True )

        info_list = dpg.enumerateGroups()

        self.assertEqual( len( info_list ), 3 )

        ids = [ x[ 'id' ] for x in info_list ]

        self.failUnless( 'everyone' in ids )
        self.failUnless( 'noone' in ids )
        self.failUnless( 'hohum' in ids )

    def test_enumerateGroups_exact_list( self ):

        from Products.PluggableAuthService.tests.test_PluggableAuthService \
            import FauxRoot

        root = FauxRoot()
        dpg = self._makeOne( 'enumerating' ).__of__( root )

        dpg.addGroup( 'everyone', 'python:True', 'Everyone', '', True )
        dpg.addGroup( 'noone', 'python:False', active=True )
        dpg.addGroup( 'hohum', 'nothing', active=True )

        ID_LIST = ( 'everyone', 'noone' )

        info_list = dpg.enumerateGroups( id=ID_LIST, exact_match=True )

        self.assertEqual( len( info_list ), len( ID_LIST ) )

        ids = [ x[ 'id' ] for x in info_list ]

        for id in ID_LIST:
            self.failUnless( id in ids )

    def test_enumerateGroups_exact_one( self ):

        from Products.PluggableAuthService.tests.test_PluggableAuthService \
            import FauxRoot

        root = FauxRoot()
        dpg = self._makeOne( 'enumerating' ).__of__( root )

        dpg.addGroup( 'everyone', 'python:True', 'Everyone', '', True )
        dpg.addGroup( 'noone', 'python:False', active=True )
        dpg.addGroup( 'hohum', 'nothing', active=True )

        info_list = dpg.enumerateGroups( id='noone', exact_match=True )

        self.assertEqual( len( info_list ), 1 )
        info = info_list[ 0 ]

        self.assertEqual( info[ 'id' ], 'noone' )
        self.assertEqual( info[ 'title' ], '' )
        self.assertEqual( info[ 'description' ], '' )
        self.assertEqual( info[ 'active' ], True )
        self.assertEqual( info[ 'predicate' ], 'python:False' )
        self.assertEqual( info[ 'pluginid' ], 'enumerating' )

        # Because teher is no proper REQUEST, the properties_url will be incorrect
        # It should normally be  '/enumerating/noone/manage_propertiesForm'
        # But it will be '//noone/manage_propertiesForm'
        URL = '//noone/manage_propertiesForm'
        self.assertEqual( info[ 'properties_url' ], URL )
        self.assertEqual( info[ 'members_url' ], URL )

    def test_enumerateGroups_exact_miss( self ):
        # See https://bugs.launchpad.net/zope-pas/+bug/585365

        from Products.PluggableAuthService.tests.test_PluggableAuthService \
            import FauxRoot

        root = FauxRoot()
        dpg = self._makeOne( 'enumerating' ).__of__( root )

        dpg.addGroup( 'everyone', 'python:True', 'Everyone', '', True )
        dpg.addGroup( 'noone', 'python:False', active=True )
        dpg.addGroup( 'hohum', 'nothing', active=True )

        info_list = dpg.enumerateGroups( id='nonesuch', exact_match=True )

        self.assertEqual( len( info_list ), 0 )

    def test_enumerateGroups_skip_inactive( self ):

        from Products.PluggableAuthService.tests.test_PluggableAuthService \
            import FauxRoot

        root = FauxRoot()
        dpg = self._makeOne( 'enumerating' ).__of__( root )

        dpg.addGroup( 'everyone', 'python:True', 'Everyone', '', True )
        dpg.addGroup( 'noone', 'python:False', active=True )
        dpg.addGroup( 'inactive', 'nothing', active=False )

        info_list = dpg.enumerateGroups()

        self.assertEqual( len( info_list ), 2 )

        ids = [ x[ 'id' ] for x in info_list ]

        self.failUnless( 'everyone' in ids )
        self.failUnless( 'noone' in ids )
        self.failIf( 'inactive' in ids )

    def test_getGroupsForPrincipal_empty( self ):

        dpg = self._makeOne( 'ggp_request' )
        principal = FauxPrincipal( 'faux' )

        groups = dpg.getGroupsForPrincipal( principal )

        self.assertEqual( len( groups ), 0 )

    def test_getGroupsForPrincipal_principal( self ):

        dpg = self._makeOne( 'ggp_principal' )
        principal = FauxPrincipal( 'faux' )

        dpg.addGroup( 'effable', 'python:principal.getId().startswith("f")' )
        groups = dpg.getGroupsForPrincipal( principal, {} )
        self.assertEqual( len( groups ), 1 )
        self.failUnless( 'effable' in groups )

    def test_getGroupsForPrincipal_python( self ):

        dpg = self._makeOne( 'ggp_python' )
        principal = FauxPrincipal( 'faux' )

        dpg.addGroup( 'everyone', 'python:1' )
        groups = dpg.getGroupsForPrincipal( principal, {} )
        self.assertEqual( len( groups ), 1 )
        self.failUnless( 'everyone' in groups )

    def test_getGroupsForPrincipal_request( self ):

        _setUpDefaultTraversable()
        dpg = self._makeOne( 'ggp_request' )
        principal = FauxPrincipal( 'faux' )

        dpg.addGroup( 'local', 'request/is_local | nothing' )

        groups = dpg.getGroupsForPrincipal( principal, {} )

        self.assertEqual( len( groups ), 0 )

        groups = dpg.getGroupsForPrincipal( principal, { 'is_local' : 0 } )

        self.assertEqual( len( groups ), 0 )

        groups = dpg.getGroupsForPrincipal( principal, { 'is_local' : 1 } )

        self.assertEqual( len( groups ), 1 )
        self.failUnless( 'local' in groups )

    def test_getGroupsForPrincipal_group( self ):

        _setUpDefaultTraversable()
        dpg = self._makeOne( 'ggp_group' )
        principal = FauxPrincipal( 'faux' )

        dpg.addGroup( 'willing', 'group/willing' )
        dpg.willing._setProperty( 'willing', type='boolean', value=0 )

        groups = dpg.getGroupsForPrincipal( principal, {} )
        self.assertEqual( len( groups ), 0 )

        dpg.willing._updateProperty( 'willing', 1 )
        groups = dpg.getGroupsForPrincipal( principal, {} )
        self.assertEqual( len( groups ), 1 )
        self.failUnless( 'willing' in groups )

    def test_getGroupsForPrincipal_plugin_nope( self ):

        dpg = self._makeOne( 'ggp_plugin' )
        principal = FauxPrincipal( 'faux' )

        dpg.addGroup( 'scripted', 'python: plugin.callme(request)' )
        callme = FauxScript( 'callme', 0 )
        dpg._setOb( 'callme', callme )

        groups = dpg.getGroupsForPrincipal( principal, {} )
        self.assertEqual( len( groups ), 0 )

    def test_getGroupsForPrincipal_plugin_ok( self ):

        dpg = self._makeOne( 'ggp_plugin' )
        principal = FauxPrincipal( 'faux' )

        dpg.addGroup( 'scripted', 'python: plugin.callme(request)' )
        callme = FauxScript( 'callme', 1 )
        dpg._setOb( 'callme', callme )

        groups = dpg.getGroupsForPrincipal( principal, {} )
        self.assertEqual( len( groups ), 1 )
        self.failUnless( 'scripted' in groups )

    def test_enumerateGroups_matching_with_optional_prefix( self ):

        from Products.PluggableAuthService.tests.test_PluggableAuthService \
            import FauxRoot

        root = FauxRoot()
        dpg = self._makeOne( 'enumerating' ).__of__( root )
        dpg.prefix = 'enumerating_'

        dpg.addGroup( 'everyone', 'python:True', 'Everyone', '', True )
        dpg.addGroup( 'noone', 'python:False', active=True )
        dpg.addGroup( 'hohum', 'nothing', active=True )

        ID_LIST = ( 'enumerating_everyone', )

        info_list = dpg.enumerateGroups( id=ID_LIST, exact_match=True )

        self.assertEqual( len( info_list ), len( ID_LIST ) )

        ids = [ x[ 'id' ] for x in info_list ]

        for id in ID_LIST:
            self.failUnless( id in ids )

    def test_enumerateGroups_enumerating_with_optional_prefix( self ):

        from Products.PluggableAuthService.tests.test_PluggableAuthService \
            import FauxRoot

        root = FauxRoot()
        dpg = self._makeOne( 'enumerating' ).__of__( root )
        dpg.prefix = 'enumerating_'

        dpg.addGroup( 'everyone', 'python:True', 'Everyone', '', True )
        dpg.addGroup( 'noone', 'python:False', active=True )
        dpg.addGroup( 'hohum', 'nothing', active=True )

        ID_LIST = ( 'enumerating_everyone', 'enumerating_noone',
                    'enumerating_hohum' )

        info_list = dpg.enumerateGroups()

        self.assertEqual( len( info_list ), len( ID_LIST ) )

        ids = [ x[ 'id' ] for x in info_list ]

        for id in ID_LIST:
            self.failUnless( id in ids )

    def test_getGroupsForPrincipal_optional_prefix( self ):

        dpg = self._makeOne( 'ggp_prefixed' )
        dpg.prefix = 'ggp_'

        principal = FauxPrincipal( 'faux' )

        dpg.addGroup( 'effable', 'python:principal.getId().startswith("f")' )
        groups = dpg.getGroupsForPrincipal( principal, {} )
        self.assertEqual( len( groups ), 1 )
        self.failUnless( 'ggp_effable' in groups )

if __name__ == "__main__":
    unittest.main()

def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite( DynamicGroupsPlugin ),
        ))



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Classes: RecursiveGroupsPlugin

$Id$
"""
from Acquisition import aq_parent
from AccessControl import ClassSecurityInfo
from App.class_init import InitializeClass
from BTrees.OOBTree import OOBTree

from zope.interface import Interface

from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from Products.PluggableAuthService.interfaces.plugins \
    import IGroupsPlugin
from Products.PluggableAuthService.PropertiedUser import PropertiedUser
from Products.PluggableAuthService.permissions import ManageGroups
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.utils import classImplements

class IRecursiveGroupsPlugin(Interface):
    """ Marker interface.
    """

manage_addRecursiveGroupsPluginForm = PageTemplateFile(
    'www/rgpAdd', globals(), __name__='manage_addRecursiveGroupsPluginForm' )

def addRecursiveGroupsPlugin( dispatcher, id, title=None, REQUEST=None ):
    """ Add a RecursiveGroupsPlugin to a Pluggable Auth Service. """

    rgp = RecursiveGroupsPlugin(id, title)
    dispatcher._setObject(rgp.getId(), rgp)

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(
                                '%s/manage_workspace'
                                '?manage_tabs_message='
                                'RecursiveGroupsPlugin+added.'
                            % dispatcher.absolute_url())

class SimpleGroup:

    def __init__( self, id ):
        self._id = id

    def getId( self ):
        return self._id

    def getGroups( self ):
        return ()

    def _addGroups( self, groups ):
        pass

class RecursiveGroupsPlugin( BasePlugin ):

    """ PAS plugin for recursively flattening a collection of groups
    """
    meta_type = 'Recursive Groups Plugin'

    security = ClassSecurityInfo()

    def __init__(self, id, title=None):

        self._id = self.id = id
        self.title = title

    #
    #   IGroupsPlugin implementation
    #
    security.declarePrivate( 'getGroupsForPrincipal' )
    def getGroupsForPrincipal( self, user, request=None ):

        set = list( user.getGroups() )
        seen = []
        parent = aq_parent( self )

        while set:
            test = set.pop(0)
            if test in seen:
                continue
            seen.append( test )
            new_groups = parent._getGroupsForPrincipal(
                PropertiedUser( test ).__of__(parent),
                ignore_plugins=( self.getId(), ) )
            if new_groups:
                set.extend( new_groups )

        return tuple( seen )

classImplements( RecursiveGroupsPlugin
               , IRecursiveGroupsPlugin
               , IGroupsPlugin
               )

InitializeClass(RecursiveGroupsPlugin)



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Class: ScriptablePlugin

$Id$
"""
from sets import Set
from urllib import quote_plus
from OFS.Folder import Folder
from AccessControl import ClassSecurityInfo
from AccessControl.Permissions import manage_users as ManageUsers
from App.class_init import default__class_init__ as InitializeClass

from zope.interface import Interface
from zope.interface import implementedBy
from zope.interface import providedBy

from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.utils import directlyProvides
from Products.PluggableAuthService.utils import classImplements

class IScriptablePlugin(Interface):
    """ Marker interface.
    """

import Products

manage_addScriptablePluginForm = PageTemplateFile(
    'www/spAdd', globals(), __name__='manage_addScriptablePluginForm' )

def addScriptablePlugin( dispatcher, id, title=None, REQUEST=None ):
    """ Add a Scriptable Plugin to a Pluggable Auth Service. """

    sp = ScriptablePlugin(id, title)
    dispatcher._setObject(sp.getId(), sp)

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(
                                '%s/manage_workspace'
                                '?manage_tabs_message='
                                'ScriptablePlugin+added.'
                            % dispatcher.absolute_url())

class ScriptablePlugin(Folder, BasePlugin):

    """ Allow users to implement plugin interfaces using script objects.

    o Allowable types include ExternalMethods, Python Scripts, ZSQL Methods,
      and DTML methods.

    o Provide UI for creating scripts for known plugin types.
    """

    security = ClassSecurityInfo()

    meta_type = 'Scriptable Plugin'

    manage_options = ( ( Folder.manage_options[0], )
                     + ( { 'label': 'Interfaces',
                           'action': 'manage_editInterfacesForm', }
                         ,
                       )
                     + BasePlugin.manage_options
                     )

    security.declareProtected( ManageUsers, 'manage_editInterfacesForm' )
    manage_editInterfacesForm = PageTemplateFile(
        'www/spEditInterfaces', globals(),
        __name__='manage_editInterfacesForm')

    def __creatable_by_emergency_user__( self ):
        return 1

    def __init__(self, id, title=None):
        self._id = self.id = id
        self.title = title

    security.declareProtected( ManageUsers, 'hasMethod')
    def hasMethod(self, method_name):
        """ Do we implement this method directly?
        """
        return method_name in self.objectIds()

    security.declarePublic('all_meta_types')
    def all_meta_types(self):
        """ What objects can be contained here? """
        allowed_types = ( 'Script (Python)'
                        , 'External Method'
                        , 'Z SQL Method'
                        , 'DTML Method'
                        , 'Page Template'
                        )

        return [x for x in Products.meta_types if x['name'] in allowed_types]

    security.declareProtected( ManageUsers, '_delOb' )
    def _delOb( self, id ):
        """
            Override ObjectManager's _delOb to account for removing any
            interface assertions the object might implement.
        """
        myId = self.getId()
        pas_instance = self._getPAS()
        plugins = pas_instance._getOb( 'plugins' )
        curr_interfaces = Set(providedBy(self))

        del_interfaces = Set([x for x in providedBy(self) if id in x.names()])

        for interface in del_interfaces:
            if myId in plugins.listPluginIds( interface ):
                plugins.deactivatePlugin( interface, myId )

        delattr( self, id )

        directlyProvides( self, *(list(curr_interfaces - del_interfaces)) )

    security.declareProtected( ManageUsers, 'manage_updateInterfaces' )
    def manage_updateInterfaces( self, interfaces, RESPONSE=None ):
        """ For ZMI update of interfaces. """

        pas_instance = self._getPAS()
        plugins = pas_instance._getOb( 'plugins' )

        new_interfaces = []

        for interface in interfaces:
            new_interfaces.append( plugins._getInterfaceFromName( interface ) )

        klass_interfaces = [x for x in implementedBy(ScriptablePlugin)]
        directlyProvides( self, *(klass_interfaces + new_interfaces) )

        if RESPONSE is not None:
            RESPONSE.redirect('%s/manage_workspace'
                              '?manage_tabs_message='
                              'Interfaces+updated.'
                            % self.absolute_url())


classImplements( ScriptablePlugin
               , IScriptablePlugin
               , *(implementedBy(Folder) + implementedBy(BasePlugin))
               )

InitializeClass(ScriptablePlugin)



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Classes: ZODBGroupManager

$Id$
"""
from Acquisition import aq_parent
from AccessControl import ClassSecurityInfo
from AccessControl.requestmethod import postonly
from App.class_init import InitializeClass
from BTrees.OOBTree import OOBTree

from zope.interface import Interface

from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from Products.PluggableAuthService.interfaces.plugins \
    import IGroupEnumerationPlugin
from Products.PluggableAuthService.interfaces.plugins \
    import IGroupsPlugin

from Products.PluggableAuthService.permissions import ManageGroups
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.utils import classImplements
from Products.PluggableAuthService.utils import csrf_only


class IZODBGroupManager(Interface):
    """ Marker interface.
    """

manage_addZODBGroupManagerForm = PageTemplateFile(
    'www/zgAdd', globals(), __name__='manage_addZODBGroupManagerForm' )

def addZODBGroupManager( dispatcher, id, title=None, REQUEST=None ):
    """ Add a ZODBGroupManager to a Pluggable Auth Service. """

    zgm = ZODBGroupManager(id, title)
    dispatcher._setObject(zgm.getId(), zgm)

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(
                                '%s/manage_workspace'
                                '?manage_tabs_message='
                                'ZODBGroupManager+added.'
                            % dispatcher.absolute_url())

class ZODBGroupManager( BasePlugin ):

    """ PAS plugin for managing groups, and groups of groups in the ZODB
    """
    meta_type = 'ZODB Group Manager'

    security = ClassSecurityInfo()

    def __init__(self, id, title=None):

        self._id = self.id = id
        self.title = title
        self._groups = OOBTree()
        self._principal_groups = OOBTree()


    #
    #   IGroupEnumerationPlugin implementation
    #
    security.declarePrivate( 'enumerateGroups' )
    def enumerateGroups( self
                        , id=None
                        , title=None
                        , exact_match=False
                        , sort_by=None
                        , max_results=None
                        , **kw
                        ):

        """ See IGroupEnumerationPlugin.
        """
        group_info = []
        group_ids = []
        plugin_id = self.getId()

        if isinstance( id, basestring ):
            id = [ id ]

        if isinstance( title, basestring ):
            title = [ title ]

        if exact_match and ( id or title ):

            if id:
                group_ids.extend( id )
            elif title:
                group_ids.extend( title )

        if group_ids:
            group_filter = None

        else:   # Searching
            group_ids = self.listGroupIds()
            group_filter = _ZODBGroupFilter( id, title, **kw )

        for group_id in group_ids:

            if self._groups.get( group_id, None ):
                e_url = '%s/manage_groups' % self.getId()
                p_qs = 'group_id=%s' % group_id
                m_qs = 'group_id=%s&assign=1' % group_id

                info = {}
                info.update( self._groups[ group_id ] )

                info[ 'pluginid' ] = plugin_id
                info[ 'properties_url' ] = '%s?%s' % ( e_url, p_qs )
                info[ 'members_url' ] = '%s?%s' % ( e_url, m_qs )

                info[ 'id' ] = '%s%s' % (self.prefix, info['id'])

                if not group_filter or group_filter( info ):
                    group_info.append( info )

        return tuple( group_info )

    #
    #   IGroupsPlugin implementation
    #
    security.declarePrivate( 'getGroupsForPrincipal' )
    def getGroupsForPrincipal( self, principal, request=None ):

        """ See IGroupsPlugin.
        """
        unadorned = self._principal_groups.get( principal.getId(), () )
        return tuple(['%s%s' % (self.prefix, x) for x in unadorned])

    #
    #   (notional)IZODBGroupManager interface
    #
    security.declareProtected( ManageGroups, 'listGroupIds' )
    def listGroupIds( self ):

        """ -> ( group_id_1, ... group_id_n )
        """
        return self._groups.keys()

    security.declareProtected( ManageGroups, 'listGroupInfo' )
    def listGroupInfo( self ):

        """ -> ( {}, ...{} )

        o Return one mapping per group, with the following keys:

          - 'id'
        """
        return self._groups.values()

    security.declareProtected( ManageGroups, 'getGroupInfo' )
    def getGroupInfo( self, group_id ):

        """ group_id -> {}
        """
        return self._groups[ group_id ]

    security.declarePrivate( 'addGroup' )
    def addGroup( self, group_id, title=None, description=None ):

        """ Add 'group_id' to the list of groups managed by this object.

        o Raise KeyError on duplicate.
        """
        if self._groups.get( group_id ) is not None:
            raise KeyError, 'Duplicate group ID: %s' % group_id

        self._groups[ group_id ] = { 'id' : group_id
                                   , 'title' : title
                                   , 'description' : description
                                   }

    security.declarePrivate( 'updateGroup' )
    def updateGroup( self, group_id, title=None, description=None ):

        """ Update properties for 'group_id'

        o Raise KeyError if group_id doesn't already exist.
        """
        if title is not None:
            self._groups[group_id]['title'] = title
        if description is not None:
            self._groups[group_id]['description'] = description
        self._groups[group_id] = self._groups[group_id]

    security.declarePrivate( 'removeGroup' )
    def removeGroup( self, group_id ):

        """ Remove 'role_id' from the list of roles managed by this
            object, removing assigned members from it before doing so.

        o Raise KeyError if 'group_id' doesn't already exist.
        """
        for principal_id in self._principal_groups.keys():
            self.removePrincipalFromGroup( principal_id, group_id )
        del self._groups[ group_id ]

    #
    #   Group assignment API
    #
    security.declareProtected( ManageGroups, 'listAvailablePrincipals' )
    def listAvailablePrincipals( self, group_id, search_id ):

        """ Return a list of principal IDs to that can belong to the group.

        o If supplied, 'search_id' constrains the principal IDs;  if not,
          return empty list.

        o Omit principals with existing assignments.
        """
        result = []

        if search_id:  # don't bother searching if no criteria

            parent = aq_parent( self )

            for info in parent.searchPrincipals( max_results=20
                                               , sort_by='id'
                                               , id=search_id
                                               , exact_match=False
                                               ):
                id = info[ 'id' ]
                title = info.get( 'title', id )
                if ( group_id not in self._principal_groups.get( id, () )
                 and group_id != id ):
                    result.append( ( id, title ) )

        return result

    security.declareProtected( ManageGroups, 'listAssignedPrincipals' )
    def listAssignedPrincipals( self, group_id ):

        """ Return a list of principal IDs belonging to a group.
        """
        result = []

        for k, v in self._principal_groups.items():
            if group_id in v:
                parent = aq_parent( self )
                info = parent.searchPrincipals( id=k, exact_match=True )
                if len( info ) == 0:
                    title = '<%s: not found>' % k
                else:
                    # always use the title of the first principal found
                    title = info[0].get( 'title', k )
                result.append( ( k, title ) )

        return result

    security.declarePrivate( 'addPrincipalToGroup' )
    def addPrincipalToGroup( self, principal_id, group_id ):

        """ Add a principal to a group.

        o Return a boolean indicating whether a new assignment was created.

        o Raise KeyError if 'group_id' is unknown.
        """
        group_info = self._groups[ group_id ] # raise KeyError if unknown!

        current = self._principal_groups.get( principal_id, () )
        already = group_id in current

        if not already:
            new = current + ( group_id, )
            self._principal_groups[ principal_id ] = new
            self._invalidatePrincipalCache( principal_id )

        return not already

    security.declarePrivate( 'removePrincipalFromGroup' )
    def removePrincipalFromGroup( self, principal_id, group_id ):

        """ Remove a prinicpal from from a group.

        o Return a boolean indicating whether the principal was already
          a member of the group.

        o Raise KeyError if 'group_id' is unknown.

        o Ignore requests to remove a principal if not already a member
          of the group.
        """
        group_info = self._groups[ group_id ] # raise KeyError if unknown!

        current = self._principal_groups.get( principal_id, () )
        new = tuple( [ x for x in current if x != group_id ] )
        already = current != new

        if already:
            self._principal_groups[ principal_id ] = new
            self._invalidatePrincipalCache( principal_id )

        return already

    #
    #   ZMI
    #
    manage_options = ( ( { 'label': 'Groups',
                           'action': 'manage_groups', }
                         ,
                       )
                     + BasePlugin.manage_options
                     )

    security.declarePublic( 'manage_widgets' )
    manage_widgets = PageTemplateFile( 'www/zuWidgets'
                                     , globals()
                                     , __name__='manage_widgets'
                                     )

    security.declareProtected( ManageGroups, 'manage_groups' )
    manage_groups = PageTemplateFile( 'www/zgGroups'
                                    , globals()
                                    , __name__='manage_groups'
                                    )

    security.declareProtected( ManageGroups, 'manage_twoLists' )
    manage_twoLists = PageTemplateFile( '../www/two_lists'
                                      , globals()
                                      , __name__='manage_twoLists'
                                      )

    security.declareProtected( ManageGroups, 'manage_addGroup' )
    def manage_addGroup( self
                       , group_id
                       , title=None
                       , description=None
                       , RESPONSE=None
                       ):
        """ Add a group via the ZMI.
        """
        self.addGroup( group_id, title, description )

        message = 'Group+added'

        if RESPONSE is not None:
            RESPONSE.redirect( '%s/manage_groups?manage_tabs_message=%s'
                             % ( self.absolute_url(), message )
                             )

    security.declareProtected( ManageGroups, 'manage_updateGroup' )
    def manage_updateGroup( self
                          , group_id
                          , title
                          , description
                          , RESPONSE=None
                          ):
        """ Update a group via the ZMI.
        """
        self.updateGroup( group_id, title, description )

        message = 'Group+updated'

        if RESPONSE is not None:
            RESPONSE.redirect( '%s/manage_groups?manage_tabs_message=%s'
                             % ( self.absolute_url(), message )
                             )

    security.declareProtected( ManageGroups, 'manage_removeGroups' )
    @csrf_only
    @postonly
    def manage_removeGroups( self
                           , group_ids
                           , RESPONSE=None
                           , REQUEST=None
                           ):
        """ Remove one or more groups via the ZMI.
        """
        group_ids = filter( None, group_ids )

        if not group_ids:
            message = 'no+groups+selected'

        else:

            for group_id in group_ids:
                self.removeGroup( group_id )

            message = 'Groups+removed'

        if RESPONSE is not None:
            RESPONSE.redirect( '%s/manage_groups?manage_tabs_message=%s'
                             % ( self.absolute_url(), message )
                             )

    security.declareProtected( ManageGroups, 'manage_addPrincipalsToGroup' )
    @csrf_only
    @postonly
    def manage_addPrincipalsToGroup( self
                                   , group_id
                                   , principal_ids
                                   , RESPONSE=None
                                   , REQUEST=None
                                   ):
        """ Add one or more principals to a group via the ZMI.
        """
        assigned = []

        for principal_id in principal_ids:
            if self.addPrincipalToGroup( principal_id, group_id ):
                assigned.append( principal_id )

        if not assigned:
            message = 'Principals+already+members+of+%s' % group_id
        else:
            message = '%s+added+to+%s' % ( '+'.join( assigned )
                                         , group_id
                                         )

        if RESPONSE is not None:
            RESPONSE.redirect( ( '%s/manage_groups?group_id=%s&assign=1'
                               + '&manage_tabs_message=%s'
                               ) % ( self.absolute_url(), group_id, message )
                             )

    security.declareProtected( ManageGroups
                             , 'manage_removePrincipalsFromGroup'
                             )
    @csrf_only
    @postonly
    def manage_removePrincipalsFromGroup( self
                                        , group_id
                                        , principal_ids
                                        , RESPONSE=None
                                        , REQUEST=None
                                        ):
        """ Remove one or more principals from a group via the ZMI.
        """
        removed = []

        for principal_id in principal_ids:
            if self.removePrincipalFromGroup( principal_id, group_id ):
                removed.append( principal_id )

        if not removed:
            message = 'Principals+not+in+group+%s' % group_id
        else:
            message = 'Principals+%s+removed+from+%s' % ( '+'.join( removed )
                                                        , group_id
                                                        )

        if RESPONSE is not None:
            RESPONSE.redirect( ( '%s/manage_groups?group_id=%s&assign=1'
                               + '&manage_tabs_message=%s'
                               ) % ( self.absolute_url(), group_id, message )
                             )

classImplements( ZODBGroupManager
               , IZODBGroupManager
               , IGroupEnumerationPlugin
               , IGroupsPlugin
               )

InitializeClass( ZODBGroupManager )

class _ZODBGroupFilter:

    def __init__( self
                , id=None
                , title=None
                , **kw
                ):

        self._filter_ids = id
        self._filter_titles = title

    def __call__( self, group_info ):

        if self._filter_ids:

            key = 'id'
            to_test = self._filter_ids

        elif self._filter_titles:

            key = 'title'
            to_test = self._filter_titles

        else:
            return 1 # TODO:  try using 'kw'

        value = group_info.get( key )

        if not value:
            return 0

        for contained in to_test:
            if value.lower().find( contained.lower() ) >= 0:
                return 1

        return 0



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" SearchPrincipalsPlugin   Plugin to delegate enumerateUsers
                             and enumerateGroups requests to another
                             PluggableAuthService
"""

# General Python imports
import copy, os
from urllib import quote_plus

# Zope imports
from Acquisition import aq_base
from OFS.Folder import Folder
from App.class_init import InitializeClass
from AccessControl import ClassSecurityInfo
from AccessControl.SpecialUsers import emergency_user

from zope.interface import Interface

from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from Products.PluggableAuthService.interfaces.plugins import \
     IUserEnumerationPlugin
from Products.PluggableAuthService.interfaces.plugins import \
     IGroupEnumerationPlugin
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.utils import classImplements

class ISearchPrincipalsPlugin(Interface):
    """ Marker interface.
    """

addSearchPrincipalsPluginForm = PageTemplateFile(
    'www/sppAdd', globals(), __name__='addSearchPrincipalsPluginForm' )


def addSearchPrincipalsPlugin( dispatcher
                             , id
                             , title=''
                             , delegate_path=''
                             , REQUEST=None
                             ):
    """ Factory method to instantiate a SearchPrincipalsPlugin """
    spp = SearchPrincipalsPlugin( id, title=title
                                , delegate_path=delegate_path )
    dispatcher._setObject(id, spp)

    if REQUEST is not None:
        REQUEST.RESPONSE.redirect('%s/manage_main' % dispatcher.absolute_url())


class SearchPrincipalsPlugin(BasePlugin):
    """ SearchPrincipalsPlugin delegates its enumerateUsers
    and enumerateGroups methods to a delegate object
    """
    security = ClassSecurityInfo()
    meta_type = 'Search Principals Plugin'

    _properties = ( { 'id' : 'delegate'
                    , 'label' : ' Delegate Path'
                    , 'type' : 'string'
                    , 'mode' : 'w'
                    }
                  ,
                  )

    def __init__(self, id, title='', delegate_path=''):
        """ Initialize a new instance """
        self.id = id
        self.title = title
        self.delegate = delegate_path

    security.declarePrivate('_getDelegate')
    def _getDelegate(self):
        """ Safely retrieve a PluggableAuthService to work with """
        uf = getattr(aq_base(self), 'acl_users', None)

        if uf is None and self.delegate:
            uf = self.unrestrictedTraverse(self.delegate)

        return uf

    security.declarePrivate('enumerateUsers')
    def enumerateUsers( self
                      , id=None
                      , login=None
                      , exact_match=0
                      , sort_by=None
                      , max_results=None
                      , **kw
                      ):
        """ see IUserEnumerationPlugin """
        acl = self._getDelegate()

        if acl is None:
            return ()

        return acl.searchUsers( id=id
                              , login=login
                              , exact_match=exact_match
                              , sort_by=sort_by
                              , max_results=max_results
                              , **kw )

    security.declarePrivate('enumerateGroups')
    def enumerateGroups( self
                       , id=None
                       , exact_match=0
                       , sort_by=None
                       , max_results=None
                       , **kw
                       ):
        """ see IGroupEnumerationPlugin """
        acl = self._getDelegate()

        if acl is None:
            return ()

        return acl.searchGroups( id=id
                               , exact_match=exact_match
                               , sort_by=sort_by
                               , max_results=max_results
                               , **kw )

classImplements( SearchPrincipalsPlugin
               , ISearchPrincipalsPlugin
               , IUserEnumerationPlugin
               , IGroupEnumerationPlugin
               )

InitializeClass(SearchPrincipalsPlugin)



##############################################################################
#
# Copyright (c) 2001 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from Products.PluggableAuthService.tests.test_PluggableAuthService \
    import FauxContainer

class FauxPAS( FauxContainer ):

    def __init__( self ):
        self._id = 'acl_users'

    def searchPrincipals( self, **kw ):
        id = kw.get( 'id' )
        return [ { 'id': id } ]

class FauxSmartPAS( FauxContainer ):

    def __init__( self ):
        self._id = 'acl_users'
        self.user_ids = {}

    def searchPrincipals( self, **kw ):
        id = kw.get( 'id' )
        prin = self.user_ids.get(id, None)
        return (prin and [ { 'id': id } ]) or []

class DummyUser:

    def __init__( self, id, groups=() ):
        self._id = id
        self._groups = groups

    def getId( self ):
        return self._id

    def getGroups( self ):
        return self._groups


def makeRequestAndResponse():
    # the POST checking requires a real HTTPRequest
    from cStringIO import StringIO
    from ZPublisher.HTTPRequest import HTTPRequest
    from ZPublisher.HTTPResponse import HTTPResponse

    res = HTTPResponse()
    req = HTTPRequest(StringIO(),
                      {'SERVER_NAME': 'localhost',
                       'SERVER_PORT': '80'},
                      res)
    return req, res



##############################################################################
#
# Copyright (c) 2001-2008 Zope Foundation and Contributors
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this
# distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import unittest

from zExceptions import Forbidden

from Products.PluggableAuthService.tests.conformance \
    import IGroupEnumerationPlugin_conformance
from Products.PluggableAuthService.tests.conformance \
    import IGroupsPlugin_conformance

from Products.PluggableAuthService.plugins.tests.helpers \
     import FauxPAS, FauxSmartPAS, DummyUser, makeRequestAndResponse

class DummyGroup:

    def __init__( self, id ):
        self._id = id

    def getId( self ):
        return self._id

class ZODBGroupManagerTests( unittest.TestCase
                           , IGroupEnumerationPlugin_conformance
                           , IGroupsPlugin_conformance
                           ):

    def _getTargetClass( self ):

        from Products.PluggableAuthService.plugins.ZODBGroupManager \
            import ZODBGroupManager

        return ZODBGroupManager

    def _makeOne( self, id='test', *args, **kw ):

        return self._getTargetClass()( id=id, *args, **kw )

    def test_empty( self ):

        zgm = self._makeOne()

        self.assertEqual( len( zgm.listGroupIds() ), 0 )
        self.assertEqual( len( zgm.enumerateGroups() ), 0 )

        user = DummyUser( 'userid' )
        groups = zgm.getGroupsForPrincipal( user )
        self.assertEqual( len( groups ), 0 )

    def test_addGroup( self ):

        from Products.PluggableAuthService.tests.test_PluggableAuthService \
            import FauxRoot

        root = FauxRoot()
        zgm = self._makeOne().__of__( root )

        zgm.addGroup( 'group' )

        group_ids = zgm.listGroupIds()
        self.assertEqual( len( group_ids ), 1 )
        self.assertEqual( group_ids[0], 'group' )

        info_list = zgm.enumerateGroups()
        self.assertEqual( len( info_list ), 1 )
        info = info_list[ 0 ]
        self.assertEqual( info[ 'id' ], 'group' )

    def test_addGroup_exists( self ):
        zgm = self._makeOne()

        zgm.addGroup( 'group' )
        self.assertRaises( KeyError, zgm.addGroup, 'group' )

    def test_updateGroup_normal( self ):
        zgm = self._makeOne()

        zgm.addGroup( 'group', 'group_title', 'group_desc' )

        group_info = zgm.getGroupInfo( 'group' )
        self.assertEqual( group_info['title'], 'group_title' )
        zgm.updateGroup( 'group', 'group_title_changed', 'group_desc' )
        group_info = zgm.getGroupInfo( 'group' )
        self.assertEqual( group_info['title'], 'group_title_changed' )

    def test_updateGroup_independent(self):
        zgm = self._makeOne()

        zgm.addGroup( 'group', 'group_title', 'group_desc' )

        zgm.updateGroup( 'group', title='group_title_changed_again')
        group_info = zgm.getGroupInfo( 'group' )
        self.assertEqual( group_info['description'], 'group_desc' )
        zgm.updateGroup( 'group', description='group_description_changed_again')
        group_info = zgm.getGroupInfo( 'group' )
        self.assertEqual( group_info['title'], 'group_title_changed_again' )


    def test_addPrincipalToGroup( self ):
        zgm = self._makeOne()

        zgm.addGroup( 'group' )

        user = DummyUser( 'userid' )

        zgm.addPrincipalToGroup( user.getId(), 'group' )
        groups = zgm.getGroupsForPrincipal( user )
        self.assertEqual( groups, ( 'group', ) )

    def test_addPrincipalToGroupThenRemovePrincipal( self ):
        root = FauxSmartPAS()
        root.user_ids['foo'] = 'foo'

        zgm = self._makeOne( id='groups' ).__of__( root )

        zgm.addGroup( 'group', 'group_title', 'group_desc' )
        self.assertEqual( len( zgm.listAssignedPrincipals( 'group' ) ), 0 )

        new = zgm.addPrincipalToGroup( 'foo', 'group' )

        self.failUnless( new )

        assigned = [x[1] for x in zgm.listAssignedPrincipals( 'group' )]

        self.assertEqual( len( assigned ), 1 )
        self.assertEqual( assigned[0], 'foo' )

        del root.user_ids['foo']

        assigned = [x[1] for x in zgm.listAssignedPrincipals( 'group' )]

        self.assertEqual( len( assigned ), 1 )
        self.assertEqual( assigned[0], '<foo: not found>' )

    def test_removePrincipalFromGroup( self ):
        zgm = self._makeOne()

        zgm.addGroup( 'group' )

        user = DummyUser( 'userid' )

        zgm.addPrincipalToGroup( user.getId(), 'group' )
        zgm.removePrincipalFromGroup( user.getId(), 'group' )
        groups = zgm.getGroupsForPrincipal( user )
        self.assertEqual( groups, () )

    def test_removeGroupOutFromUnderPrincipal( self ):
        zgm = self._makeOne()

        zgm.addGroup( 'group' )

        user = DummyUser( 'userid' )

        zgm.addPrincipalToGroup( user.getId(), 'group' )
        zgm.removeGroup( 'group' )
        group_ids = zgm.listGroupIds()
        self.assertEqual( len( group_ids ), 0 )
        groups = zgm.getGroupsForPrincipal( user )
        self.assertEqual( groups, () )

    def test_multiplePrincipalsPerGroup( self ):
        pas = FauxPAS()
        zgm = self._makeOne().__of__( pas )

        zgm.addGroup( 'group1' )
        zgm.addGroup( 'group2' )

        user1 = DummyUser( 'userid1' )
        user2 = DummyUser( 'userid2' )

        zgm.addPrincipalToGroup( user1.getId(), 'group1' )
        zgm.addPrincipalToGroup( user1.getId(), 'group2' )
        zgm.addPrincipalToGroup( user2.getId(), 'group2' )

        group_ids = zgm.listGroupIds()
        self.assertEqual( len( group_ids ), 2 )
        principals = zgm.listAssignedPrincipals( 'group2' )
        self.assertEqual( principals, [ ( 'userid1', 'userid1' ),
                                        ( 'userid2', 'userid2' ) ] )

    def test_enumerateGroups_exact_nonesuch( self ):
        from Products.PluggableAuthService.tests.test_PluggableAuthService \
            import FauxRoot

        root = FauxRoot()
        zgm = self._makeOne( id='exact_nonesuch' ).__of__( root )

        ID_LIST = ( 'foo', 'bar', 'baz', 'bam' )

        for id in ID_LIST:

            zgm.addGroup( id, 'Group %s' % id, 'This is group, %s' % id )

        self.assertEqual( zgm.enumerateGroups( id='qux', exact_match=True )
                        , () )

    def test_enumerateGroups_exact_string( self ):
        from Products.PluggableAuthService.tests.test_PluggableAuthService \
            import FauxRoot

        root = FauxRoot()
        zgm = self._makeOne(id='exact_nonesuch').__of__(root)

        ID_LIST = ('foo', 'bar', 'baz', 'bam')

        for id in ID_LIST:
            zgm.addGroup(id, 'Group %s' % id, 'This is group, %s' % id)

        info = zgm.enumerateGroups(id='foo', exact_match=True)
        self.assertEqual(len(info), 1)
        self.assertEqual(info[0]['id'], 'foo')

    def test_enumerateGroups_exact_unicode( self ):
        from Products.PluggableAuthService.tests.test_PluggableAuthService \
            import FauxRoot

        root = FauxRoot()
        zgm = self._makeOne(id='exact_nonesuch').__of__(root)

        ID_LIST = ('foo', 'bar', 'baz', 'bam')

        for id in ID_LIST:
            zgm.addGroup(id, 'Group %s' % id, 'This is group, %s' % id)

        info = zgm.enumerateGroups(id=u'foo', exact_match=True)
        self.assertEqual(len(info), 1)
        self.assertEqual(info[0]['id'], 'foo')

    def test_enumerateGroups_multiple( self ):
        from Products.PluggableAuthService.tests.test_PluggableAuthService \
            import FauxRoot

        root = FauxRoot()
        zrm = self._makeOne( id='partial' ).__of__( root )

        ID_LIST = ( 'foo', 'bar', 'baz', 'bam' )

        for id in ID_LIST:
            zrm.addGroup( id, 'Group %s' % id, 'This is group, %s' % id )

        info_list = zrm.enumerateGroups( id=ID_LIST, exact_match=False )

        self.assertEqual( len( info_list ), len( ID_LIST ) )

        for info in info_list:
            self.failUnless( info[ 'id' ] in ID_LIST )

        SUBSET = ID_LIST[:3]

        info_list = zrm.enumerateGroups( id=SUBSET, exact_match=False )

        self.assertEqual( len( info_list ), len( SUBSET ) )

        for info in info_list:
            self.failUnless( info[ 'id' ] in SUBSET )

    def test_enumerateGroups_prefixed( self ):
        from Products.PluggableAuthService.tests.test_PluggableAuthService \
            import FauxRoot

        root = FauxRoot()
        zrm = self._makeOne( id='prefixed' ).__of__( root )
        zrm.prefix = 'prefixed_'

        ID_LIST = ( 'foo', 'bar', 'baz', 'bam' )
        PRE_LIST = tuple(['prefixed_%s' % x for x in ID_LIST])

        for id in ID_LIST:

            zrm.addGroup( id, 'Group %s' % id, 'This is group, %s' % id )

        info_list = zrm.enumerateGroups()

        self.assertEqual( len( info_list ), len( ID_LIST ) )

        for info in info_list:
            self.failUnless( info[ 'id' ] in PRE_LIST )

    def test_addPrincipalToGroup_w_prefix( self ):
        zgm = self._makeOne()
        zgm.prefix = 'prefixed_'

        zgm.addGroup( 'group' )

        user = DummyUser( 'userid' )

        zgm.addPrincipalToGroup( user.getId(), 'group' )
        groups = zgm.getGroupsForPrincipal( user )
        self.assertEqual( groups, ( 'prefixed_group', ) )

    def test_manage_addPrincipalsToGroup_POST_permissions(self):
        USER_ID = 'testuser'
        GROUP_ID = 'testgroup'

        zgm = self._makeOne()
        zgm.prefix = 'prefixed_'

        zgm.addGroup(GROUP_ID)
        req, res = makeRequestAndResponse()

        req.set('REQUEST_METHOD', 'GET')
        req.set('method', 'GET')
        req.set('SESSION', {})
        self.assertRaises(Forbidden, zgm.manage_addPrincipalsToGroup,
                          GROUP_ID, [USER_ID], REQUEST=req)

        req.set('REQUEST_METHOD', 'POST')
        req.set('method', 'POST')
        self.assertRaises(Forbidden, zgm.manage_addPrincipalsToGroup,
                          GROUP_ID, [USER_ID], REQUEST=req)

        # Works with a POST + CSRF token
        req.form['csrf_token'] = 'deadbeef'
        req.SESSION['_csrft_'] = 'deadbeef'
        zgm.manage_addPrincipalsToGroup(GROUP_ID, [USER_ID], REQUEST=req)

    def test_manage_removePrincipalsFromGroup_POST_permissions(self):
        USER_ID = 'testuser'
        GROUP_ID = 'testgroup'

        zgm = self._makeOne()
        zgm.prefix = 'prefixed_'

        zgm.addGroup(GROUP_ID)
        req, res = makeRequestAndResponse()

        req.set('REQUEST_METHOD', 'GET')
        req.set('method', 'GET')
        req.set('SESSION', {})
        self.assertRaises(Forbidden, zgm.manage_removePrincipalsFromGroup,
                          GROUP_ID, [USER_ID], REQUEST=req)

        # Works with a POST
        req.set('REQUEST_METHOD', 'POST')
        req.set('method', 'POST')
        self.assertRaises(Forbidden, zgm.manage_removePrincipalsFromGroup,
                          GROUP_ID, [USER_ID], REQUEST=req)

        # Works with a POST + CSRF token
        req.form['csrf_token'] = 'deadbeef'
        req.SESSION['_csrft_'] = 'deadbeef'
        zgm.manage_removePrincipalsFromGroup(GROUP_ID, [USER_ID], REQUEST=req)

    def test_manage_removeGroup_POST_permissions(self):
        GROUP_ID = 'testgroup'

        zgm = self._makeOne()
        zgm.prefix = 'prefixed_'
        zgm.addGroup(GROUP_ID)
        req, res = makeRequestAndResponse()

        req.set('REQUEST_METHOD', 'GET')
        req.set('method', 'GET')
        req.set('SESSION', {})
        self.assertRaises(Forbidden, zgm.manage_removeGroups,
                          [GROUP_ID], REQUEST=req)

        req.set('REQUEST_METHOD', 'POST')
        req.set('method', 'POST')
        self.assertRaises(Forbidden, zgm.manage_removeGroups,
                          [GROUP_ID], REQUEST=req)

        # Works with a POST + CSRF token
        req.form['csrf_token'] = 'deadbeef'
        req.SESSION['_csrft_'] = 'deadbeef'
        zgm.manage_removeGroups([GROUP_ID], REQUEST=req)


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite( ZODBGroupManagerTests ),
        ))
